# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule contains functions to plot densities of states and
bandstructures for electronic and vibrational calculations.

"""


import os
import copy
import numpy as np
from matador.utils.viz_utils import get_element_colours
from matador.plotting.plotting import plotting_function
from matador.scrapers import optados2dict, phonon2dict, bands2dict, phonon_dos2dict
from matador.scrapers import cell2dict, res2dict
from matador.orm.spectral import (
    ElectronicDispersion, ElectronicDOS,
    VibrationalDispersion, VibrationalDOS,
    Dispersion, DensityOfStates
)
from matador.utils.chem_utils import INVERSE_CM_TO_EV

__all__ = ['plot_spectral']


@plotting_function
def plot_spectral(seeds, **kwargs):
    """ This function wraps all of the spectral plotting capability of matador.
    When provided with a seed, or seeds, several files will be checked:

        - <seed>.bands: CASTEP bandstructure expected, not DOS,
        - <seed>.adaptive.dat: OptaDOS total DOS,
        - <seed>.pdos.dat: OptaDOS pDOS,
        - <seed>.pdis.dat: OptaDOS projected dispersion,

    or, if the "phonons" flag is passed, or if a <seed>.phonon file is detected,

        - <seed>.phonon: CASTEP phonon dispersion curves expected,
        - <seed>.phonon_dos: CASTEP phonon DOS.

    This function will then automatically choose what to plot, favouring a bandstructure
    with "the-most-projected" DOS it can find.

    Parameters:
        seeds (list): list of filenames of bands/phonon files

    Keyword Arguments:
        plot_bandstructure (bool): whether to plot bandstructure, if available
        plot_dos (bool): whether to plot density of states, if available
        plot_pdos (bool): whether or not to plot projected DOS, if available
        plot_pdis (bool): whether to plot projected dispersion, if available
        dos (str): separate seed name for pDOS/DOS data
        phonons (bool): whether to plot phonon or electronic data
        labels (list): list of strings for legend labels for multiple bandstructures
        gap (bool): whether to draw on the band gap
        colour_by_seed (bool): plot with a separate colour per bandstructure
        external_efermi (float or list): replace scraped Fermi energy with this value (eV) (can be
            specified per spin channel).
        highlight_bands (list): list of integer indices, colour the bands
            with these indices in red
        band_colour (str): if passed "occ", bands will be coloured using
            cmap depending on whether they lie above or below the Fermi
            level. Otherwise, override all colour options with
            matplotlib-interpretable colour (e.g. hexcode or html colour
            name) to use for all bands (DEFAULT: 'occ').
        band_alpha (float): transparency of plotted bands.
        filename (str): filename for figure saving.
        cmap (str): matplotlib colourmap name to use for the bands
        cmap_limits (tuple): fraction of cmap to use (DEFAULT: (0.2, 0.8)).
        n_colours (int): number of colours to use from cmap (DEFAULT: 4).
        unstacked_pdos (bool): whether to plot projected DOS as stack or overlapping.
        spin_only (str): either 'up' or 'down' to only plot one spin channel.
        preserve_kspace_distance (bool): whether to preserve distances in reciprocal space when
            linearising the kpoint path. If False, bandstructures of different lattice parameters
            with the same Bravais lattice can be more easily compared. If True, bandstructures may
            appear rarefied or compressed in particular regions.
        pdis_interpolation_factor (float): multiple by which to interpolate pDIS bands
        pdis_point_scale (float): size of points in pDIS (DEFAULT: 50).
        projectors_to_plot (str): comma-separted list of projectors to
            plot in the PDIS or PDOS, provided as element:orbital, e.g.
            "K:s,K:p,P" will plot s and p orbitals for K, and all orbitals for P.
        band_reorder (bool): try to reorder bands based on local gradients (DEFAULT: True for phonons, otherwise False).
        title (str): optional plot title
        pdos_hide_sum (bool): whether or not to plot the total DOS on a PDOS plot; this is to hide
            regions where the PDOS is negative (leading to total DOS lower than stacked PDOS) (DEFAULT: False).

    """
    import matplotlib.pyplot as plt
    from cycler import cycler
    # set defaults and update class with desired values
    prop_defaults = {'plot_bandstructure': True, 'plot_dos': True, 'plot_pdos': True, 'plot_pdis': True,
                     'phonons': False, 'gap': False,
                     'colour_by_seed': False, 'external_efermi': None,
                     'labels': None, 'cmap': None, 'cmap_limits': (0.2, 0.8), 'band_colour': 'occ',
                     'n_colours': 4, 'spin_only': None, 'figsize': None, 'filename': None,
                     'pdis_interpolation_factor': 2, 'pdis_point_scale': 25, 'projectors_to_plot': None,
                     'unstacked_pdos': False, 'preserve_kspace_distance': False,
                     'band_reorder': False, 'title': None, 'show': True,
                     'verbosity': 0, 'highlight_bands': None, 'pdos_hide_sum': True}

    for key in kwargs:
        if kwargs[key] is not None:
            prop_defaults[key] = kwargs[key]
    kwargs = prop_defaults

    kwargs['projectors_to_plot'] = _parse_projectors_list(kwargs['projectors_to_plot'])

    if kwargs.get('cmap') is None:
        kwargs['colours'] = list(plt.rcParams['axes.prop_cycle'].by_key()['color'])
        plt.rcParams['axes.prop_cycle'] = cycler('color', kwargs['colours'])
    else:
        if isinstance(kwargs['cmap'], str):
            print('Adjusting colour palette... to {}'.format(kwargs.get('cmap')))
            try:
                kwargs['colours'] = plt.cm.get_cmap(kwargs.get('cmap')).colors
                plt.rcParams['axes.prop_cycle'] = cycler('color', kwargs['colours'])
            except AttributeError:
                kwargs['colours'] = list(plt.rcParams['axes.prop_cycle'].by_key()['color'])
            kwargs['_mpl_cmap'] = plt.get_cmap(kwargs.get('cmap'))
        elif isinstance(kwargs['cmap'], list):
            print('Reading list of colours {}...'.format(kwargs.get('cmap')))
            kwargs['colours'] = kwargs['cmap']
            plt.rcParams['axes.prop_cycle'] = cycler('color', kwargs['colours'])

    if (kwargs.get('phonons') and kwargs['band_colour'] == 'occ' and kwargs.get('cmap') is None):
        kwargs['band_colour'] = 'grey'
        if kwargs.get('band_alpha') is None:
            kwargs['band_alpha'] = 0.8

    elif kwargs['band_colour'] in [None, 'random']:
        kwargs['band_colour'] = None

    if not isinstance(seeds, list):
        seeds = [seeds]

    if len(seeds) > 1:
        if kwargs['plot_pdis'] or kwargs['plot_dos']:
            kwargs['plot_pdos'] = False
            kwargs['plot_pdis'] = False
            print('Disabling projections as mutiple seeds requested.')

    if kwargs.get('plot_window') is not None:
        if isinstance(kwargs.get('plot_window'), list):
            if len(kwargs.get('plot_window')) != 2:
                exit('plot_window must have length 2 or be a single number')
            kwargs['plot_window'] = sorted(kwargs.get('plot_window'))
        else:
            kwargs['plot_window'] = (-kwargs.get('plot_window'), kwargs.get('plot_window'))
    else:
        kwargs['plot_window'] = None

    if all(isinstance(seed, str) for seed in seeds):

        if kwargs['plot_dos']:
            # check an optados file exists
            exts = ['pdos.dat', 'adaptive.dat', 'fixed.dat', 'linear.dat', 'jdos.dat', 'phonon_dos', 'bands_dos']
            kwargs['plot_dos'] = any([any([os.path.isfile('{}.{}'.format(seed, ext)) for ext in exts]) for seed in seeds])

        if kwargs['plot_pdos']:
            exts = ['pdos.dat', 'phonon_dos']
            kwargs['plot_pdos'] = any([any([os.path.isfile('{}.{}'.format(seed, ext)) for ext in exts]) for seed in seeds])

    figsize = kwargs['figsize']
    if kwargs['plot_bandstructure'] and not kwargs['plot_dos']:
        if figsize is None:
            figsize = (7, 6)
        fig, ax_dispersion = plt.subplots(figsize=figsize)
    elif kwargs['plot_bandstructure'] and kwargs['plot_dos']:
        if figsize is None:
            figsize = (10, 6)
        fig, ax_grid = plt.subplots(1, 3, figsize=figsize, sharey=True,
                                    gridspec_kw={'width_ratios': [4, 2, 1],
                                                 'wspace': 0.1,
                                                 'left': 0.15})
        ax_dispersion = ax_grid[0]
        ax_dos = ax_grid[1]
        ax_grid[2].axis('off')
    elif not kwargs['plot_bandstructure'] and kwargs['plot_dos']:
        if figsize is None:
            figsize = (9, 4)
        fig, ax_dos = plt.subplots(1, figsize=figsize)
    else:
        raise RuntimeError("No plots requested, please set either plot_dos or plot_bandstructure to True!")

    kwargs['valence'] = kwargs['colours'][0]
    kwargs['conduction'] = kwargs['colours'][-1]
    kwargs['crossing'] = kwargs['colours'][int(len(kwargs['colours']) / 2)]

    if len(seeds) > 1 or kwargs.get('colour_by_seed'):
        kwargs['seed_colours'] = kwargs['colours']
        kwargs['ls'] = ['-'] * len(seeds)
        kwargs['colour_by_seed'] = True
        if kwargs.get('labels') is None:
            try:
                kwargs['labels'] = [seed.split('/')[-1].split('.')[0] for seed in seeds]
            except AttributeError:
                kwargs['labels'] = [seed.root_source for seed in seeds]

    kwargs['ls'] = []
    for i in range(len(seeds)):
        if i % 3 == 0:
            kwargs['ls'].append('-')
        elif i % 3 == 1:
            kwargs['ls'].append('--')
        elif i % 3 == 2:
            kwargs['ls'].append('-.')

    bbox_extra_artists = []
    if kwargs['plot_bandstructure']:
        ax_dispersion = dispersion_plot(seeds, ax_dispersion, kwargs, bbox_extra_artists)

    if kwargs['plot_dos']:
        ax_dos = dos_plot(seeds, ax_dos, kwargs, bbox_extra_artists)

    if kwargs.get('title') is not None:
        fig.suptitle(kwargs.get('title'))

    if any([kwargs.get('pdf'), kwargs.get('svg'), kwargs.get('png')]):
        if not bbox_extra_artists:
            bbox_extra_artists = None
        filename = kwargs.get('filename')
        if filename is None:
            filename = seeds[0].split('/')[-1].replace('.bands', '').replace('.phonon', '') + '_spectral'
        if kwargs.get('pdf'):
            plt.savefig('{}.pdf'.format(filename),
                        bbox_inches='tight', transparent=True, bbox_extra_artists=bbox_extra_artists)
        if kwargs.get('svg'):
            plt.savefig('{}.svg'.format(filename),
                        bbox_inches='tight', transparent=True, bbox_extra_artists=bbox_extra_artists)
        if kwargs.get('png'):
            plt.savefig('{}.png'.format(filename),
                        bbox_inches='tight', transparent=True, bbox_extra_artists=bbox_extra_artists)

    else:
        plt.tight_layout()
        if kwargs['show']:
            print('Displaying plot...')
            plt.show()


def dispersion_plot(seeds, ax_dispersion, kwargs, bbox_extra_artists):
    """ Plot a dispersion/bandstructure on the given axis. Will detect
    and projected dispersion data automatically.

    Parameters:
        seeds (list): the seednames of the data to plot.
        ax_dispersion (matplotlib.Axes): the axis to plot on.
        kwargs (dict): any plotting keywords (from e.g. dispersion script).
        bbox_extra_artists (list): a list to which to append legends.

    Returns:
        matplotlib.Axes: the axis that was plotted on.

    """
    from cycler import cycler
    import matplotlib.pyplot as plt
    plotted_pdis = False
    for seed_ind, seed in enumerate(seeds):

        if isinstance(seed, dict):
            if kwargs.get('phonons'):
                dispersion = VibrationalDispersion(seed)
            else:
                dispersion = ElectronicDispersion(seed)

        elif isinstance(seed, Dispersion):
            dispersion = seed

        else:
            seed = seed.replace('.bands', '').replace('.phonon', '')
            if os.path.isfile('{}.phonon'.format(seed)):
                dispersion, s = phonon2dict(seed + '.phonon', verbosity=kwargs.get('verbosity'))
                if not s:
                    raise RuntimeError(dispersion)

                dispersion = VibrationalDispersion(dispersion)

            elif os.path.isfile('{}.bands'.format(seed)):
                dispersion, s = bands2dict(seed + '.bands',
                                           verbosity=kwargs.get('verbosity'))
                if not s:
                    raise RuntimeError(dispersion)

                if os.path.isfile('{}.pdis.dat'.format(seed)) and kwargs['plot_pdis']:
                    pdis_data, s = optados2dict('{}.pdis.dat'.format(seed))
                    if not s:
                        raise RuntimeError(pdis_data)
                else:
                    pdis_data = None

                dispersion = ElectronicDispersion(dispersion, projection_data=pdis_data)

            else:
                raise RuntimeError('{}.bands/.phonon not found.'.format(seed))

        eigs = np.array(dispersion.eigs, copy=True)
        if kwargs['phonons']:
            # convert from internal eV frequencies to cm^-1
            eigs /= INVERSE_CM_TO_EV

        if kwargs['plot_window'] is None:
            if kwargs['phonons']:
                kwargs['plot_window'] = [min(-10, np.min(eigs) - 10), np.max(eigs)]
            else:
                kwargs['plot_window'] = [-10, 10]

        # try to match bands if requested
        if kwargs['band_reorder']:
            print('Reordering bands based on local gradients...')
            eigs = Dispersion.get_band_reordering(eigs, dispersion.kpoint_branches)

        path = dispersion.linearise_path(preserve_kspace_distance=kwargs['preserve_kspace_distance'])

        if dispersion.projectors and len(seeds) == 1 and kwargs['plot_pdis'] and not kwargs['phonons']:
            ax_dispersion = projected_bandstructure_plot(dispersion, ax_dispersion, path,
                                                         bbox_extra_artists, eigs=eigs,
                                                         **kwargs)
            kwargs['band_colour'] = 'grey'
            plotted_pdis = True

        # loop over branches and plot
        if not plotted_pdis:
            if kwargs.get('external_efermi') is None:
                spin_fermi_energy = dispersion.spin_fermi_energy
            else:
                spin_fermi_energy = kwargs.get('external_efermi')
            if len(spin_fermi_energy) == 1 and dispersion.num_spins != 1:
                spin_fermi_energy = [spin_fermi_energy] * dispersion.num_spins

            if kwargs.get('cmap') is not None:
                cmap_limits = kwargs.get('cmap_limits', (0.2, 0.8))
                kwargs['_mpl_cmap'] = plt.cm.get_cmap(kwargs.get('cmap'))(
                    np.linspace(*cmap_limits, num=dispersion.num_bands)
                )

            # loop over branches and plot
            for branch_ind, branch in enumerate(dispersion.kpoint_branches):
                plt.rcParams['axes.prop_cycle'] = cycler('color', kwargs['colours'])

                for ns in range(dispersion.num_spins):
                    if ns == 1 and kwargs.get('spin_only') == 'up':
                        continue
                    elif ns == 0 and kwargs.get('spin_only') == 'down':
                        continue

                    for nb in range(dispersion.num_bands):
                        colour, alpha, label = _get_lineprops(
                            dispersion, spin_fermi_energy, nb, ns, branch, branch_ind, seed_ind, kwargs, eigs=eigs)

                        ax_dispersion.plot(path[(np.asarray(branch)-branch_ind).tolist()],
                                           eigs[ns][nb][branch] - spin_fermi_energy[ns],
                                           c=colour, ls=kwargs['ls'][seed_ind], alpha=alpha, label=label)

    if len(seeds) > 1:
        disp_legend = ax_dispersion.legend(loc='upper center')
        bbox_extra_artists.append(disp_legend)

    ax_dispersion.axhline(0, ls='--', lw=1, c='grey')
    ax_dispersion.set_ylim(kwargs['plot_window'])
    if kwargs['phonons']:
        ylabel = 'Wavenumber (cm$^{-1}$)'
    else:
        ylabel = 'Energy (eV)'
    ax_dispersion.set_ylabel(ylabel)
    ax_dispersion.set_xlim(0, 1)
    _add_path_labels(seeds[-1], dispersion, ax_dispersion, path, 0, kwargs)

    return ax_dispersion


def dos_plot(seeds, ax_dos, kwargs, bbox_extra_artists):
    """ Plot a density of states on the given axis. Will detect
    pDOS and spin-dependent DOS data automatically.

    Parameters:
        seeds (list): the seednames of the data to plot.
        ax_dos (matplotlib.Axes): the axis to plot on.
        kwargs (dict): any plotting keywords (from e.g. dispersion script).
        bbox_extra_artists (list): a list to which to append legends.

    Returns:
        matplotlib.Axes: the axis that was plotted on.

    """
    for seed_ind, seed in enumerate(seeds):
        # load electronic data
        if not kwargs['phonons']:
            dos_data = _load_electronic_dos(seed, kwargs)

            if kwargs['plot_window'] is None:
                kwargs['plot_window'] = [-10, 10]

            if kwargs['plot_pdos']:
                if 'pdos' in dos_data:
                    pdos_data = dos_data
                else:
                    pdos_seed = '{}.pdos.dat'.format(seed)
                    if os.path.isfile(pdos_seed):
                        pdos_data, s = optados2dict(pdos_seed, verbosity=0)
                        if not s:
                            raise RuntimeError(pdos_data)
                        dos_data['pdos'] = pdos_data
        else:
            dos_data = _load_phonon_dos(seed, kwargs)
            max_density = np.max(dos_data['dos'])
            if kwargs['plot_pdos']:
                pdos_data = dos_data

        energies = np.copy(dos_data['energies'])
        # change unit of phonon energies and set plot window
        if kwargs['phonons']:
            energies /= INVERSE_CM_TO_EV
            if kwargs['plot_window'] is None:
                kwargs['plot_window'] = [np.min(energies[np.where(dos_data['dos'] > 1e-3)]) - 10,
                                         np.max(energies[np.where(dos_data['dos'] > 1e-3)])]

        if 'spin_dos' in dos_data:
            max_density = max(np.max(np.abs(dos_data['spin_dos']['down'][np.where(energies > kwargs['plot_window'][0])])),
                              np.max(np.abs(dos_data['spin_dos']['up'][np.where(energies > kwargs['plot_window'][0])])))
        else:
            max_density = np.max(dos_data['dos'][np.where(np.logical_and(energies < kwargs['plot_window'][1],
                                                                         energies > kwargs['plot_window'][0]))])

        dos = dos_data['dos']

        # plotting pdos depends on these other factors too
        plotting_pdos = (kwargs['plot_pdos'] and len(seeds) == 1
                         and not (kwargs['phonons'] and len(dos_data.get('pdos', [])) <= 1))

        if kwargs['phonons']:
            ylabel = 'Phonon DOS'
            xlabel = 'Wavenumber (cm$^{{-1}}$)'
        else:
            if 'dos_unit_label' in dos_data:
                ylabel = dos_data['dos_unit_label'].replace('A^3', 'Å$^{3}$')
            else:
                if kwargs['plot_bandstructure']:
                    ylabel = 'DOS'
                else:
                    ylabel = 'DOS (eV$^{{-1}}$Å$^{{-3}}$)'
            xlabel = 'Energy (eV)'

        if kwargs['plot_bandstructure']:
            ax_dos.set_xlabel(ylabel)
            ax_dos.axhline(0, c='grey', ls='--', lw=1)
            if 'spin_dos' in dos_data:
                ax_dos.set_xlim(-max_density*1.2, max_density * 1.2)
            else:
                ax_dos.set_xlim(0, max_density * 1.2)
            ax_dos.set_ylim(kwargs['plot_window'])
            ax_dos.axvline(0, c='grey', lw=1)
            ax_dos.xaxis.set_ticks_position('none')

            if 'spin_dos' not in dos_data:
                ax_dos.plot(dos, energies, ls=kwargs['ls'][seed_ind],
                            color='grey', zorder=1e10, label='Total DOS')
                if not kwargs['plot_pdos']:
                    ax_dos.fill_betweenx(energies[np.where(energies > 0)], 0, dos[np.where(energies > 0)],
                                         alpha=0.2, color=kwargs['conduction'])
                    ax_dos.fill_betweenx(energies[np.where(energies <= 0)], 0, dos[np.where(energies <= 0)],
                                         alpha=0.2, color=kwargs['valence'])
        else:
            ax_dos.set_xlabel(xlabel)
            ax_dos.set_ylabel(ylabel)
            ax_dos.axvline(0, c='grey', lw=1, ls='--')
            if 'spin_dos' in dos_data:
                ax_dos.set_ylim(-max_density*1.2, max_density * 1.2)
            else:
                ax_dos.set_ylim(0, max_density * 1.2)
            ax_dos.set_xlim(kwargs['plot_window'])
            ax_dos.axhline(0, c='grey', lw=1)

            if 'spin_dos' not in dos_data:
                ax_dos.plot(energies, dos, ls=kwargs['ls'][seed_ind], alpha=1,
                            c='grey', zorder=1e10, label='Total DOS')
                if not plotting_pdos:
                    ax_dos.fill_between(energies[np.where(energies > 0)], 0, dos[np.where(energies > 0)],
                                        alpha=0.2, color=kwargs['conduction'])
                    ax_dos.fill_between(energies[np.where(energies <= 0)], 0, dos[np.where(energies <= 0)],
                                        alpha=0.2, color=kwargs['valence'])

        if 'spin_dos' in dos_data and not kwargs['pdos_hide_sum']:
            if kwargs['plot_bandstructure']:
                if kwargs.get('spin_only') in [None, 'up']:
                    if not plotting_pdos:
                        ax_dos.fill_betweenx(energies, 0, dos_data['spin_dos']['up'], alpha=0.2, color='r')
                    ax_dos.plot(dos_data['spin_dos']['up'], energies,
                                ls=kwargs['ls'][seed_ind], color='r', zorder=1e10, label='$\\uparrow$')
                if kwargs.get('spin_only') in [None, 'down']:
                    if not plotting_pdos:
                        ax_dos.fill_betweenx(energies, 0, -dos_data['spin_dos']['down'], alpha=0.2, color='b')
                    ax_dos.plot(-dos_data['spin_dos']['down'], energies,
                                ls=kwargs['ls'][seed_ind], color='b', zorder=1e10, label='$\\downarrow$')
            else:
                if kwargs.get('spin_only') in [None, 'up']:
                    ax_dos.plot(energies, dos_data['spin_dos']['up'],
                                ls=kwargs['ls'][seed_ind], color='r', zorder=1e10, label='$\\uparrow$')
                    if not plotting_pdos:
                        ax_dos.fill_between(energies, 0, dos_data['spin_dos']['up'], alpha=0.2, color='r')
                if kwargs.get('spin_only') in [None, 'down']:
                    ax_dos.plot(energies, -dos_data['spin_dos']['down'],
                                ls=kwargs['ls'][seed_ind], color='b', zorder=1e10, label='$\\downarrow$')
                    if not plotting_pdos:
                        ax_dos.fill_between(energies, 0, -dos_data['spin_dos']['down'], alpha=0.2, color='b')

        if plotting_pdos:

            if kwargs.get('projectors_to_plot') is not None:
                pdos = dict()
                for projector in pdos_data['pdos']:
                    if projector in kwargs.get("projectors_to_plot"):
                        pdos[projector] = pdos_data['pdos'][projector]
            else:
                pdos = pdos_data['pdos']

            stacks = dict()
            projector_labels, dos_colours = _get_projector_info(list(pdos.keys()))
            unique_labels = set()
            for ind, projector in enumerate(pdos):

                # don't break PDOS label down by spin
                if projector_labels[ind] in unique_labels:
                    projector_labels[ind] = ''
                else:
                    unique_labels.add(projector_labels[ind])

                # split stacked pdos by spin channel
                stack_key = None
                if len(projector) > 2:
                    stack_key = projector[2]

                if stack_key not in stacks:
                    stacks[stack_key] = np.zeros_like(pdos[projector])

                stack = stacks[stack_key]
                if kwargs['unstacked_pdos']:
                    stack = 0
                else:
                    stack = stacks[stack_key]

                if not kwargs['unstacked_pdos']:
                    alpha = 0.8
                else:
                    alpha = 0.7

                # mask negative contributions with 0
                pdos[projector] = np.ma.masked_where(pdos[projector] < 0, pdos[projector], copy=True)
                np.ma.set_fill_value(pdos[projector], 0)
                pdos[projector] = np.ma.filled(pdos[projector])

                # flip sign of down spin energies for spin polarised plot
                if 'down' in projector:
                    pdos[projector] *= -1

                if not np.max(np.abs(pdos[projector])) < 1e-8:
                    if kwargs['plot_bandstructure']:
                        label = None
                        if not kwargs['unstacked_pdos']:
                            ax_dos.fill_betweenx(energies, stack, stack+pdos[projector],
                                                 alpha=alpha, label=projector_labels[ind],
                                                 color=dos_colours[ind])
                        else:
                            label = projector_labels[ind]
                        ax_dos.plot(stack + pdos[projector], energies,
                                    alpha=1, color=dos_colours[ind], label=label)
                    else:
                        label = None
                        if not kwargs['unstacked_pdos']:
                            ax_dos.fill_between(energies, stack, stack+pdos[projector],
                                                alpha=alpha, label=projector_labels[ind],
                                                color=dos_colours[ind])
                        else:
                            label = projector_labels[ind]
                        ax_dos.plot(energies, stack + pdos[projector],
                                    alpha=1, color=dos_colours[ind], label=label)

                    stacks[stack_key] += pdos[projector]

            if not kwargs['pdos_hide_sum'] and kwargs['unstacked_pdos']:
                for stack_key in stacks:
                    if stack_key is None:
                        label = 'Sum pDOS'
                    else:
                        label = 'Sum pDOS: spin-{}'.format(stack_key)
                    if kwargs['plot_bandstructure']:
                        ax_dos.plot(stacks[stack_key], energies,
                                    ls='--', alpha=1, color='black', zorder=1e9, label=label)
                    else:
                        ax_dos.plot(energies, stacks[stack_key],
                                    ls='--', alpha=1, color='black', zorder=1e9, label=label)

        if len(seeds) == 1 and (plotting_pdos or 'spin_dos' in dos_data):
            if kwargs['plot_bandstructure']:
                dos_legend = ax_dos.legend(bbox_to_anchor=(1, 1))

            else:
                dos_legend = ax_dos.legend(bbox_to_anchor=(1, 0.5), loc='center left')

            bbox_extra_artists.append(dos_legend)

    return ax_dos


def projected_bandstructure_plot(
    dispersion,
    ax,
    path,
    bbox_extra_artists,
    eigs=None,
    pdis_interpolation_factor=2,
    pdis_point_scale=25,
    projectors_to_plot=None,
    **kwargs
):
    """ Plot projected bandstructure with weightings from OptaDOS pdis.dat file.

    Parameters:
        dispersion (matador.orm.spectral.ElectronicDispersion): scraped
            data for bandstructure and pdis.
        seed (str): seed name for files to scrape.
        ax (matplotlib.pyplot.Axes): axis to plot on.
        bbox_extra_artists (list): list to append any legends too.

    Keyword arguments:
        eigs (np.ndarray): eigenvalues for the associated Dispesion object,
            passed separately to allow for reordering.
        interpolation_factor (float): amount by which to interpolate bands.
        point_scale (float): rescale points by this amount
        projectors_to_plot (list(tuple)): list of projectors to plot.

    Returns:
        matplotlib.pyplot.Axes: the axis that was plotted on.

    """

    if eigs is None:
        eigs = dispersion.eigs_s_k

    if projectors_to_plot is not None:
        if not any(projector in dispersion.projectors for projector in projectors_to_plot):
            raise RuntimeError(
                "None of the desired projectors {} could be found in {}"
                .format(projectors_to_plot, dispersion.projectors)
            )

        _projectors_to_plot = []
        _projector_inds = []
        for ind, projector in enumerate(dispersion.projectors):
            if projector in projectors_to_plot:
                _projectors_to_plot.append(projector)
                _projector_inds.append(ind)

        pdis = np.zeros((dispersion.num_kpoints, dispersion.num_bands, len(_projectors_to_plot)))
        for jnd, ind in enumerate(_projector_inds):
            pdis[:, :, jnd] = dispersion.projector_weights[:, :, ind]

        projectors = _projectors_to_plot

    else:
        pdis = np.array(dispersion.projector_weights, copy=True)
        projectors = copy.deepcopy(dispersion.projectors)

    pdis[pdis < 0] = 0
    pdis[pdis > 1] = 1

    keep_inds = []
    for ind, _ in enumerate(projectors):
        if np.max(pdis[:, :, ind]) > 1e-8:
            keep_inds.append(ind)

    projector_labels, dos_colours = _get_projector_info(projectors)

    fermi_energy = kwargs.get('external_efermi') or dispersion.fermi_energy

    _ordered_scatter(path, eigs[0].T - fermi_energy, pdis, dispersion.kpoint_branches,
                     interpolation_factor=pdis_interpolation_factor, point_scale=pdis_point_scale,
                     ax=ax, colours=dos_colours)

    for ind, _ in enumerate(projectors):
        if ind in keep_inds:
            ax.scatter(1e20, 0, facecolor=dos_colours[ind],
                       label=projector_labels[ind], lw=0)

    legend = ax.legend(loc=1)
    legend.set_zorder(1e20)
    bbox_extra_artists.append(legend)

    return ax


def _ordered_scatter(path, eigs, pdis, branches, ax=None, colours=None, interpolation_factor=2, point_scale=25):
    """ Plots an ordered scatter plot of a projected bandstructure.

    Parameters:
        path (np.ndarray): linearised [0, 1] kpoint path array.
        eigs (np.ndarray): (num_kpoints x num_bands) array containing eigenvalues
        pdis (np.ndarray): (num_kpoints x num_bands x num_projectors) array containing
            projector weights.
        branches (list): list of branch indices, e.g. for two branches [[0,1,2], [3, 4]].

    Keyword arguments:
        ax (matplotlib.Axes): axis to plot on
        colours (list): colours assigned for each projector.
        interpolation_factor (float): multiplier for fineness of band interpolation.

    """
    from scipy.interpolate import interp1d
    flat_pts_k = []
    flat_pts_e = []
    flat_sizes = []
    flat_colours = []
    flat_zorders = []

    for nb in range(len(eigs[0])):
        for branch_ind, branch in enumerate(branches):
            k = path[(np.asarray(branch) - branch_ind).tolist()]
            projections = pdis[branch, nb]
            ek_fn = interp1d(k, eigs[branch, nb])
            k_interp = np.linspace(np.min(k), np.max(k), num=int(interpolation_factor*len(k)))
            ek_interp = ek_fn(k_interp)
            projections = projections.T
            interp_projections = []
            for i, _ in enumerate(projections):
                interp_projections.append(interp1d(k, projections[i])(k_interp))
            projections = np.asarray(interp_projections).T
            pts = np.array([k_interp, ek_interp]).T.reshape(-1, 1, 2)

            if colours is not None:
                plot_colours = [colours[i] for i in range(len(projections[0]))]
            else:
                plot_colours = [None for i in range(len(projections[0]))]
            for i, _ in enumerate(projections):
                # zeros mess up zorder, so add small shift then subtract
                # before calculating real sizes
                projections[projections <= 1e-9] = 1e-9
                sizes = np.cumsum(projections[i])
                zorders = 1000*(-nb + 1-sizes)
                projections[projections <= 1e-9] = 0
                sizes = np.cumsum(projections[i])
                for j in range(len(projections[i])):
                    flat_pts_k.append(pts[i, 0, 0])
                    flat_pts_e.append(pts[i, 0, 1])
                    size = sizes[j]
                    flat_sizes.append(point_scale*(size)**2)
                    flat_colours.append(plot_colours[j])
                    flat_zorders.append(zorders[j])
            ax.plot(pts[:, 0, 0], pts[:, 0, 1], lw=0.5, alpha=0.5, c='grey', zorder=0)

    flat_zorders = np.asarray(flat_zorders)
    flat_pts_k = np.asarray(flat_pts_k)[np.argsort(flat_zorders)]
    flat_pts_e = np.asarray(flat_pts_e)[np.argsort(flat_zorders)]
    flat_sizes = np.asarray(flat_sizes)[np.argsort(flat_zorders)]
    flat_colours = np.asarray(flat_colours)[np.argsort(flat_zorders)]

    ax.scatter(flat_pts_k, flat_pts_e, edgecolor=flat_colours, s=flat_sizes, lw=0, marker='o', facecolor=flat_colours)


def _get_lineprops(dispersion, spin_fermi_energy, nb, ns, branch, branch_ind, seed_ind, kwargs, eigs=None):
    """ Get the properties of the line to plot. """
    colour = None
    alpha = 1
    label = None

    if isinstance(dispersion, ElectronicDispersion):
        if eigs is None:
            eigs = dispersion.eigs
        if dispersion.num_spins == 2:
            if ns == 0:
                colour = 'red'
                alpha = 0.8
            else:
                colour = 'blue'
                alpha = 0.8
        else:
            if kwargs.get('band_colour') == 'occ':
                band_min = np.min(eigs[ns][nb][branch]) - spin_fermi_energy[ns]
                band_max = np.max(eigs[ns][nb][branch]) - spin_fermi_energy[ns]

                if band_max < 0:
                    colour = kwargs.get('valence')
                elif band_min > 0:
                    colour = kwargs.get('conduction')
                elif band_min < 0 < band_max:
                    colour = kwargs.get('crossing')

    if kwargs['colour_by_seed']:
        colour = kwargs.get('seed_colours')[seed_ind]

    if kwargs.get('band_colour') is not None:
        if kwargs.get('band_colour') != 'occ':
            colour = kwargs.get('band_colour')

    if kwargs.get('_mpl_cmap') is not None:
        colour = kwargs['_mpl_cmap'][nb]

    if kwargs.get('band_alpha') is not None:
        alpha = kwargs['band_alpha']

    if kwargs.get('highlight_bands') is not None:
        if nb in kwargs.get('highlight_bands'):
            colour = 'red'
        else:
            alpha = 0.5

    if branch_ind == 0 and ns == 0 and nb == 0 and kwargs.get('labels') is not None:
        label = kwargs.get('labels')[seed_ind]

    return colour, alpha, label


def _add_path_labels(seed, dispersion, ax_dispersion, path, seed_ind, kwargs):
    """ Scrape k-point path labels from cell file and seekpath, then add them to the plot. """
    from matador.utils.cell_utils import doc2spg, get_seekpath_kpoint_path
    xticks = []
    xticklabels = []
    shear_planes = []
    labelled = []
    path_labels = dict()

    # first, try to grab them from the cell file
    if isinstance(seed, str) and os.path.isfile(seed + '.cell'):
        doc, success = cell2dict(seed + '.cell',
                                 db=False, verbosity=kwargs.get('verbosity', 0),
                                 lattice=True, positions=True)
    else:
        doc = seed

    if kwargs['phonons']:
        key = 'phonon_fine_kpoint_path'
    else:
        key = 'spectral_kpoints_path'
    if key in doc and key + '_labels' in doc:
        for label, point in zip(doc.get(key + '_labels', []), doc.get(key, None)):
            path_labels[label] = point
        print('Detected path labels from cell file')

    if not path_labels:
        # try to get dispersion path labels from spglib/seekpath
        spg_structure = None
        if isinstance(dispersion, Dispersion):
            try:
                spg_structure = doc2spg(dispersion)
            except (KeyError, RuntimeError) as exc:
                print(f"Unable to create spglib structure from input data: skipping path labels: {exc}.")

        if not spg_structure:
            res = False
            cell = False
            if isinstance(seed, str):
                if os.path.isfile(seed + '.res'):
                    res = True
                elif os.path.isfile(seed + '.cell'):
                    cell = True
                else:
                    print('Failed to find {}.cell or {}.res, will not be able to generate labels.'.format(seed, seed))

            success = False
            if cell:
                doc, success = cell2dict(seed + '.cell',
                                         db=False, verbosity=kwargs.get('verbosity', 0),
                                         lattice=True, positions=True)
            if res and not success:
                doc, success = res2dict(seed + '.res',
                                        db=False, verbosity=kwargs.get('verbosity', 0))

            if cell or res:
                if success:
                    spg_structure = doc2spg(doc)
                else:
                    print('Failed to scrape {}.cell/.res, will not be able to generate labels.'.format(seed))

        if spg_structure:
            _, _, seekpath_results = get_seekpath_kpoint_path(spg_structure, standardize=False, explicit=False)
            path_labels = seekpath_results['point_coords']

    for branch_ind, branch in enumerate(dispersion.kpoint_branches):
        for sub_ind, ind in enumerate(branch):
            kpt = dispersion.kpoint_path[ind]
            for label, point in path_labels.items():
                if np.allclose(point, kpt):
                    if ind - branch_ind not in labelled:
                        label = label.replace('GAMMA', r'\Gamma')
                        label = label.replace('SIGMA', r'\Sigma')
                        label = label.replace('DELTA', r'\Delta')
                        label = label.replace('LAMBDA', r'\Lambda')
                        if sub_ind == len(branch) - 1:
                            if branch_ind < len(dispersion.kpoint_branches) - 1:
                                _tmp = dispersion.kpoint_path
                                next_point = _tmp[dispersion.kpoint_branches[branch_ind + 1][0]]
                                for new_label, new_point in path_labels.items():
                                    new_label = new_label.replace('GAMMA', r'\Gamma')
                                    new_label = new_label.replace('SIGMA', r'\Sigma')
                                    new_label = new_label.replace('DELTA', r'\Delta')
                                    new_label = new_label.replace('LAMBDA', r'\Lambda')
                                    # import matplotlib
                                    if np.allclose(new_point, next_point):
                                        label = '\\dfrac{{{}}}{{{}}}'.format(label, new_label)
                                        ax_dispersion.axvline(path[ind - branch_ind], ls='-', c='grey', zorder=1, lw=0.5)
                                        labelled.append(ind - branch_ind)
                                        shear_planes.append(ind)
                        label = '${}$'.format(label.replace('$', ''))
                        ax_dispersion.axvline(path[ind - branch_ind], ls='--', c='grey', zorder=0, lw=0.5)
                        xticklabels.append(label)
                        xticks.append(path[ind - branch_ind])
                        break

    if isinstance(dispersion, ElectronicDispersion) and kwargs['gap']:
        if dispersion.num_spins != 1:
            raise NotImplementedError('Band gap summary not implemented for multiple spin channels.')
        if dispersion.band_gap > 0:
            vbm_pos = dispersion['band_gap_path_inds'][1]
            vbm = dispersion['valence_band_min'] - dispersion.fermi_energy
            cbm_pos = dispersion['band_gap_path_inds'][0]
            cbm = dispersion['conduction_band_max'] - dispersion.fermi_energy
            if vbm_pos != cbm_pos:
                vbm_offset = sum([vbm_pos > ind for ind in shear_planes])
                cbm_offset = sum([cbm_pos > ind for ind in shear_planes])
                ax_dispersion.plot([path[vbm_pos - vbm_offset], path[cbm_pos - cbm_offset]], [vbm, cbm],
                                   ls=kwargs['ls'][seed_ind],
                                   c='blue',
                                   label='indirect gap {:3.3f} eV'.format(cbm - vbm))

            vbm_pos = dispersion['direct_gap_path_inds'][1]
            vbm = dispersion['direct_valence_band_min'] - dispersion.fermi_energy
            cbm_pos = dispersion['direct_gap_path_inds'][0]
            cbm = dispersion['direct_conduction_band_max'] - dispersion.fermi_energy
            vbm_offset = sum([vbm_pos > ind for ind in shear_planes])
            cbm_offset = sum([cbm_pos > ind for ind in shear_planes])
            ax_dispersion.plot([path[vbm_pos - vbm_offset], path[cbm_pos - cbm_offset]], [vbm, cbm],
                               ls=kwargs['ls'][seed_ind],
                               c='red',
                               label='direct gap {:3.3f} eV'.format(cbm - vbm))
            ax_dispersion.legend(loc='upper center',
                                 bbox_to_anchor=(0.5, 1.1),
                                 fancybox=True, shadow=True,
                                 ncol=2, handlelength=1)

    if seed_ind == 0:
        ax_dispersion.set_xticks(xticks)
        ax_dispersion.set_xticklabels(xticklabels)
        ax_dispersion.grid(False)


def _get_projector_info(projectors):
    """ Grab appropriate colours and labels from a list of projectors.

    Parameters:
        projectors (list): list containing (element_str, l_channel) tuples.

    Returns:
        list: list of projector labels, e.g. {element_str}-${l_channel}$.
        list: list of colours for density of states, derived from vesta colours.

    """

    import matplotlib.pyplot as plt
    element_colours = get_element_colours()
    projector_labels = []
    dos_colours = []
    for ind, projector in enumerate(projectors):

        # pad out projectors for e.g. phonon case
        species = projector[0]
        if len(projector) > 1:
            ang_mom = projector[1]
        else:
            ang_mom = None
        if len(projector) > 2:
            spin = projector[2]
        else:
            spin = None

        # (species, None, None)
        if species is not None and ang_mom is None and spin is None:
            projector_label = species
        # (None, ang mom, None)
        if species is None and ang_mom is not None and spin is None:
            projector_label = '${}$'.format(ang_mom)
        # (None, None, spin)
        elif species is None and ang_mom is None and spin is not None:
            projector_label = ''
        # (species, ang_mom, None/spin)
        elif species is not None and ang_mom is not None:
            projector_label = '{} (${}$)'.format(species, ang_mom)
        # (species, None, None/spin)
        elif species is not None and ang_mom is None:
            projector_label = '{}'.format(species)
        # (None, ang_mom, None/spin)
        elif species is None and ang_mom is not None:
            projector_label = '${}$'.format(ang_mom)
        # (species, ang_mom, None/spin)
        else:
            projector_label = '{} (${}$)'.format(species, ang_mom)

        projector_labels.append(projector_label)

        # if species-projected only, then use VESTA colours
        if species is not None and ang_mom is None:
            dos_colours.append(element_colours.get(projector[0]))
        # if species_ang-projected, then use VESTA colours but lightened
        elif species is not None and ang_mom is not None:
            dos_colour = copy.deepcopy(element_colours.get(projector[0]))
            multi = ['s', 'p', 'd', 'f'].index(projector[1]) - 1
            for jind, _ in enumerate(dos_colour):
                dos_colour[jind] = max(min(dos_colour[jind]+multi*0.2, 1), 0)
            dos_colours.append(dos_colour)
        # otherwise if just ang-projected, use colour_cycle
        if dos_colours[-1] is None:
            dos_colours[-1] = list(plt.rcParams['axes.prop_cycle'].by_key()['color'])[ind]

    return projector_labels, dos_colours


def _load_electronic_dos(seed, kwargs):
    """ Try to obtain electronic DOS data, either from files, or as
    a dictionary.

    Parameters:
        seed (str/dict): either a filename or dictionary containing dos
            data.
        kwargs (dict): plotting kwargs.

    Returns:
        ElectronicDOS object containing scraped data.

    """
    if isinstance(seed, dict):
        return ElectronicDOS(seed)

    if isinstance(seed, ElectronicDOS):
        return seed

    seed = seed.replace('.bands', '')
    if kwargs.get('dos') is None:
        # look for dat files, and just use the first
        exts = ['adaptive.dat', 'fixed.dat', 'linear.dat', 'bands_dos']
        for ext in exts:
            if os.path.isfile('{}.{}'.format(seed, ext)):
                dos_seed = '{}.{}'.format(seed, ext)
                break
        else:
            raise RuntimeError('No total DOS files found.')
    else:
        dos_seed = kwargs.get('dos')

    # If bands_dos exists, do some manual broadening:
    # .bands_dos is a file written by run3 when doing a
    # full spectral calculation, it is simply the .bands
    # file output from a DOS calculation
    if dos_seed.endswith('.bands_dos'):
        dos_data, s = bands2dict(dos_seed)
        gaussian_width = kwargs.get('gaussian_width', 0.1)
        dos_data['dos'], dos_data['energies'] = DensityOfStates.bands_as_dos(
            dos_data, gaussian_width=gaussian_width
        )

        if isinstance(dos_data['dos'], dict):
            dos_data['spin_dos'] = dos_data['dos']
            del dos_data['dos']

    else:
        dos_data, s = optados2dict(dos_seed, verbosity=0)

    if not s:
        raise RuntimeError(dos_data)

    return ElectronicDOS(dos_data)


def _load_phonon_dos(seed, kwargs):
    """ Try to obtain phonon DOS data, either from files, or as
    a dictionary.

    Parameters:
        seed (str/dict): either a filename or dictionary containing dos
            data.
        kwargs (dict): plotting kwargs.

    Returns:
        VibrationalDOS object containing scraped data.

    """

    if isinstance(seed, dict):
        return VibrationalDOS(seed)
    if isinstance(seed, DensityOfStates):
        return seed

    # otherwise, just read the phonon_dos file
    dos_data, s = phonon_dos2dict(seed + '.phonon_dos')
    if not s:
        raise RuntimeError(dos_data)

    return VibrationalDOS(dos_data)


def _parse_projectors_list(projectors):
    """ Convert CLI args into the appropriate projector, ignoring
    spin channels.

    Parameters:
        projectors (str): a string of comma-separated element:orbital
            pairs. If the colon is omitted, all oribtals will be used.

    Returns:
        list(tuple): list of projectors in format [(element, orbital, spin)].

    """
    if projectors is None:
        return None
    _projectors = []
    orbitals = ['s', 'p', 'd', 'f']

    for projector in projectors.split(','):
        if ":" not in projector:
            element = projector
            for orbital in orbitals:
                _projectors.append((element, orbital, None))
            _projectors.append((element, None, None))
        else:
            element = projector.split(":")[0]
            orbital = projector.split(":")[1]
            _projectors.append((element, orbital, None))

    return _projectors



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements some light wrappers to
the pymatgen, via ASE.

"""

import pickle
import os
import getpass

from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen import MPRester
from matador.utils.ase_utils import ase2dict, doc2ase
from matador.utils.cell_utils import calc_mp_spacing

REQ_PROPERTIES = [
    'formula',
    'input',
    'doi',
    'e_above_hull',
    'snl',
    'material_id',
    'icsd_ids',
    'pf_ids',
    'formation_energy_per_atom',
    'structure'
]


def get_chemsys(elements, dumpfile=None):
    """ Scrape the Materials Project for the chemical system specified
    by elements, e.g. for elements `['A', 'B', 'C']` the query performed
    is `chemsys='A-B-C' & nelements=3`. Requires interactive user input
    of their MP API key (unless set by environment variable
    $PMG_MAPI_KEY).


    Parameters:
        elements (list): list of chemical symbols.

    Keyword arguments:
        dumpfile (str): optional filename to dump the pickled response
            after conversion to matador documents.

    """
    cursor = []
    count = {}
    api_key = os.environ.get('PMG_MAPI_KEY', None)
    if api_key is None:
        api_key = getpass.getpass('MP API key:').strip()

    with MPRester(api_key=api_key) as m:
        data = {
            'criteria': {
                'chemsys': '-'.join(elements),
                'nelements': len(elements),
            },
            'properties': REQ_PROPERTIES
        }
        response = m.query(**data)
        for result in response:
            cursor.append(mp2dict(result))

    if dumpfile:
        with open(dumpfile, 'wb') as f:
            pickle.dump(cursor, f)

    print('ICSD structures: {}'.format(sum(1 for doc in cursor if 'icsd' in doc)))
    print('PF structures: {}'.format(sum(1 for doc in cursor if '_mp_pf_ids' in doc)))

    return cursor, count


def doc2pmg(doc):
    """ Converts matador document to a pymatgen structure,
    via ASE.

    """
    ase_atoms = doc2ase(doc)
    return AseAtomsAdaptor.get_structure(ase_atoms)


def mp2dict(response):
    """ Convert a response from pymatgen.MPRester into a matador document,
    via an ASE atoms object. Expects certain properties to be requested
    in order to construct a full matador document, e.g. `structure` & `input`.

    Parameters:
        response (dict): containing one item of the MPRester response.

    """

    if 'structure' not in response:
        raise RuntimeError('`structure` key missing, nothing to scrape...')

    ase_struct = AseAtomsAdaptor.get_atoms(response['structure'])
    doc = ase2dict(ase_struct)

    doc['source'] = []

    if 'material_id' in response:
        doc['_mp_id'] = response['material_id']
        doc['source'].append('{}'.format(doc['_mp_id']))

    if 'pf_ids' in response:
        if response['pf_ids']:
            doc['_mp_pf_ids'] = response['pf_ids']
            for _id in doc['_mp_pf_ids']:
                doc['source'].append('pf-{}'.format(_id))

    if 'icsd_ids' in response:
        if response['icsd_ids']:
            doc['icsd'] = response['icsd_ids'][0]
        doc['_mp_icsd_ids'] = response['icsd_ids']

    if 'formation_energy_per_atom' in response:
        doc['formation_energy_per_atom'] = response['formation_energy_per_atom']
        doc['enthalpy_per_atom'] = response['formation_energy_per_atom']
        doc['enthalpy'] = doc['enthalpy_per_atom'] * doc['num_atoms']
        doc['formation_energy'] = doc['formation_energy_per_atom'] * doc['num_atoms']

    if 'e_above_hull' in response:
        doc['hull_distance'] = response['e_above_hull']

    if 'doi' in response:
        doc['_mp_doi'] = response['doi']

    if 'snl' in response:
        doc['_mp_references'] = response['snl'].references

    if 'input' in response:
        incar = response['input']['incar']
        if not isinstance(incar, dict):
            incar = incar.as_dict()
        doc['cut_off_energy'] = incar['ENCUT']
        doc['xc_functional'] = 'PBE'
        kpts = response['input']['kpoints']
        doc['kpoints_mp_grid'] = kpts.kpts[0]
        doc['kpoints_mp_offset'] = kpts.kpts_shift
        doc['pressure'] = 0.0
        doc['stress'] = 0.0
        doc['kpoints_mp_spacing'] = calc_mp_spacing(doc['lattice_cart'], doc['kpoints_mp_grid'])
        doc['spin_polarized'] = bool(incar['ISPIN'] - 1)

    return doc



""" This submodule contains aliases for some useful constants. """

from scipy.constants import physical_constants

FARADAY_CONSTANT_Cpermol = physical_constants['Faraday constant'][0]
HARTREE_TO_EV = physical_constants['Hartree energy in eV'][0]
Cperg_to_mAhperg = 2.778e-1
C_TO_mAh = Cperg_to_mAhperg
BOHR_TO_ANGSTROM = physical_constants['Bohr radius'][0] * 1e10
RY_TO_EV = physical_constants['Rydberg constant times hc in eV'][0]
KBAR_TO_GPA = 0.1
eV_PER_ANGSTROM_CUBED_TO_GPa = 160.21776
AVOGADROS_NUMBER = physical_constants['Avogadro constant'][0]
ANGSTROM_CUBED_TO_CENTIMETRE_CUBED = 1e-24
ELECTRON_CHARGE = physical_constants['elementary charge'][0]
KELVIN_TO_EV = physical_constants['kelvin-electron volt relationship'][0]
INVERSE_CM_TO_EV = physical_constants['inverse meter-electron volt relationship'][0] * 100

__all__ = [
    'FARADAY_CONSTANT_Cpermol',
    'HARTREE_TO_EV',
    'Cperg_to_mAhperg',
    'C_TO_mAh',
    'BOHR_TO_ANGSTROM',
    'RY_TO_EV',
    'KBAR_TO_GPA',
    'eV_PER_ANGSTROM_CUBED_TO_GPa',
    'AVOGADROS_NUMBER',
    'ANGSTROM_CUBED_TO_CENTIMETRE_CUBED',
    'ELECTRON_CHARGE',
    'KELVIN_TO_EV',
    'INVERSE_CM_TO_EV',
]



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule implements the Crystal class, a wrapper
to the raw dictionary stored in MongoDB that allows for validation,
manipulation and analysis of the lattice.

"""

from copy import deepcopy
from matador.utils import cell_utils
from matador.orm.orm import DataContainer
from matador.crystal.crystal_site import Site
from matador.utils.chem_utils import get_concentration
from matador.utils.cell_utils import real2recip


class UnitCell:
    """ This class describes a 3D periodic unit cell by its
    Cartesian lattice vectors or lattice parameters, in Å.

    """

    _lattice_abc = None
    _lattice_cart = None
    _volume = None

    def __init__(self, lattice):
        """ Initialise the cell from either Cartesian lattice vectors
        [[x1, y1, z1], [x2, y2, z2], [x3, y3, z3]], or lattice
        parameters [[a, b, c], [alpha, beta, gamma]].

        Parameters:
            lattice (:obj:`list` or numpy.ndarray): either Cartesian
                lattice vectors or lattice parameters, stored as lists
                or a numpy arrays.

        """

        self._volume = None
        if len(lattice) == 3:
            if any(len(vec) != 3 for vec in lattice):
                raise RuntimeError('Unable to cast {} into lattice_cart'.format(lattice))
            self.lattice_cart = lattice
        elif len(lattice) == 2:
            if any(len(vec) != 3 for vec in lattice):
                raise RuntimeError('Unable to cast {} into lattice_abc'.format(lattice))
            self.lattice_abc = lattice
        else:
            raise RuntimeError("Unable to create UnitCell from lattice {}".format(lattice))

    @property
    def lattice_cart(self):
        """ The Cartesian lattice vectors as a tuple. """
        return self._lattice_cart

    @lattice_cart.setter
    def lattice_cart(self, new_lattice):
        self._lattice_cart = tuple(tuple(vec) for vec in new_lattice)
        self._lattice_abc = tuple(tuple(elem) for elem in cell_utils.cart2abc(self._lattice_cart))
        self._volume = None

    @property
    def lattice_abc(self):
        """ Lattice parameters as a tuple. """
        return self._lattice_abc

    @lattice_abc.setter
    def lattice_abc(self, new_lattice):
        self._lattice_abc = tuple(tuple(vec) for vec in new_lattice)
        self._lattice_cart = tuple(tuple(vec) for vec in cell_utils.abc2cart(self._lattice_abc))
        self._volume = None

    @property
    def lengths(self):
        """ Lattice vector lengths. """
        return self._lattice_abc[0]

    @lengths.setter
    def lengths(self, new_lengths):
        if len(new_lengths) != 3:
            raise RuntimeError('Expected list of 3 floats for cell vector lengths, received {}'
                               .format(new_lengths))
        self.lattice_abc = [new_lengths, self._lattice_abc[1]]

    @property
    def recip_lattice_cart(self):
        return real2recip(self.lattice_cart)

    @property
    def angles(self):
        """ Lattice vector angles. """
        return self._lattice_abc[1]

    @angles.setter
    def angles(self, new_angles):
        if len(new_angles) != 3:
            raise RuntimeError('Expected list of 3 floats for cell vector angles, received {}'
                               .format(new_angles))
        self.lattice_abc = [self._lattice_abc[0], new_angles]

    @property
    def volume(self):
        """ The cell volume in Å³. """
        if not self._volume:
            self._volume = cell_utils.cart2volume(self._lattice_cart)
        return self._volume


class Crystal(DataContainer):
    """ Class that wraps the MongoDB document, providing useful
    interfaces for cell manipulation and validation.

    Attributes:
        elems (:obj:`list` of :obj:`str`): list of present elements in the crystal,
            sorted in alphabetical order.
        cell (:obj:`UnitCell`): the unit cell constructed from lattice vectors.

    """

    @staticmethod
    def _validate_doc(doc):
        if not any(key in doc for key in ['lattice_cart', 'lattice_abc']):
            raise RuntimeError('No lattice information found, cannot create Crystal.')
        if 'atom_types' not in doc:
            raise RuntimeError('No species information found `"atom_types"`, cannot create Crystal.')
        if not any(key in doc for key in ['positions_frac', 'positions_abs']):
            raise RuntimeError('No position information found `"positions_frac"/"positions_abs"`, cannot create Crystal.')

    def __init__(self, doc, voronoi=False, network_kwargs=None):
        """ Initialise Crystal object from matador document with Site list
        and any additional abstractions, e.g. voronoi or CrystalGraph.

        Parameters:
            doc (dict): document containing structural information, minimal requirement
                is for `atom_types`, one of `lattice_abc` or `lattice_cart`, and one of
                `positions_frac` or `positions_abs` to be present.

        Keyword Arguments:
           voronoi (bool): whether to compute Voronoi substructure for each site
           network_kwargs (dict): keywords to pass to the CrystalGraph initialiser

        """

        self._validate_doc(doc)
        if isinstance(doc, Crystal):
            doc = deepcopy(doc._data)

        super().__init__(doc)

        self.elems = sorted(list(set(self._data['atom_types'])))
        self.sites = []

        # use lattice_cart to construct cell if present, otherwise abc
        self.cell = UnitCell(doc.get('lattice_cart', doc.get('lattice_abc')))

        self._construct_sites(voronoi=voronoi)

        if network_kwargs is not None:
            self._network_kwargs = network_kwargs
        else:
            self._network_kwargs = {}

        # assign attributes for later
        self._coordination_lists = None
        self._coordination_stats = None
        self._network = None
        self._bond_lengths = None
        self._bonding_stats = None

        # assume default value for symprec
        if 'space_group' in self._data:
            self._space_group = {0.01: self._data['space_group']}
        else:
            self._space_group = {}

    def __getitem__(self, key):
        """ If integer key is requested, return index into site array. """
        if isinstance(key, int):
            return self.sites[key]
        return super().__getitem__(key)

    def __str__(self):
        repr_string = "{root_source}: {formula}\n".format(root_source=self.root_source, formula=self.formula)
        repr_string += "{num_atoms:<3} atoms. {space_group:<8}\n".format(num_atoms=self.num_atoms,
                                                                         space_group=self.space_group)

        if 'formation_enthalpy_per_atom' in self._data:
            repr_string += ("Formation enthalpy = {:6.6f} eV/atom\n".format(self._data['formation_enthalpy_per_atom']))

        repr_string += (
            "(a, b, c) = {lattice[0][0]:4.4f} Å, {lattice[0][1]:4.4f} Å, {lattice[0][2]:4.4f} Å\n"
            "(α, β, γ) = {lattice[1][0]:4.4f}° {lattice[1][1]:4.4f}° {lattice[1][2]:4.4f}°\n"
            .format(lattice=self.cell.lattice_abc))

        return repr_string

    def set_positions(self, new_positions, fractional=True):
        if len(new_positions) != self.num_atoms:
            raise RuntimeError('Cannot change size of positions array!')
        if fractional:
            self._data['positions_frac'] = new_positions
            self._data.pop('positions_abs', None)
        else:
            self._data['positions_abs'] = new_positions
            self._data.pop('positions_frac', None)

        self._construct_sites()

    def _construct_sites(self, voronoi=False):
        """ Constructs the list of Site objects stored in self.sites.

        Keyword arguments:
            voronoi (bool): whether to calculate the Voronoi substructure
                of each site.

        """
        self.sites = []
        for ind, species in enumerate(self.atom_types):
            position = self.positions_frac[ind]
            site_data = {}
            if 'site_occupancy' in self._data:
                if len(self._data['site_occupancy']) == len(self._data['atom_types']):
                    site_data['site_occupancy'] = self._data['site_occupancy'][ind]
            if 'chemical_shifts' in self._data:
                if len(self._data['chemical_shifts']) == len(self._data['atom_types']):
                    site_data['magres_shift'] = self._data['chemical_shifts'][ind]
            if 'magnetic_shielding_tensor' in self._data:
                if len(self._data['magnetic_shielding_tensor']) == len(self._data['atom_types']):
                    site_data['magres_shielding'] = self._data['magnetic_shielding_tensor'][ind]
            if 'chemical_shift_anisos' in self._data:
                site_data['magres_aniso'] = self._data['chemical_shift_anisos'][ind]
            if 'chemical_shift_asymmetries' in self._data:
                site_data['magres_asymm'] = self._data['chemical_shift_asymmetries'][ind]
            if 'atomic_spins' in self._data:
                site_data['spin'] = self._data['atomic_spins'][ind]
            if 'voronoi_substructure' in self._data:
                site_data['voronoi_substructure'] = self._data['voronoi_substructure'][ind]
            elif voronoi:
                site_data['voronoi_substructure'] = self.voronoi_substructure[ind]

            self.sites.append(Site(species, position, self.cell.lattice_cart, **site_data))

    @property
    def atom_types(self):
        """ Return list of atom types. """
        return self._data['atom_types']

    @property
    def num_atoms(self):
        """ Return number of atoms in structure. """
        return len(self.sites)

    @property
    def num_elements(self):
        """ Return number of species in the structure. """
        return len(self.elems)

    @property
    def positions_frac(self):
        """ Return list of fractional positions. """
        from matador.utils.cell_utils import cart2frac
        if 'positions_frac' not in self._data:
            self._data['positions_frac'] = cart2frac(self.cell.lattice_cart, self.positions_abs)
        return self._data['positions_frac']

    @property
    def positions_abs(self):
        """ Return list of absolute Cartesian positions. """
        from matador.utils.cell_utils import frac2cart
        if 'positions_abs' not in self._data:
            self._data['positions_abs'] = frac2cart(self.cell.lattice_cart, self.positions_frac)
        return self._data['positions_abs']

    @property
    def site_occupancies(self):
        """ Return a list of site occupancies. """
        if 'site_occupancies' not in self._data:
            self._data['site_occupancies'] = [site.occupancy for site in self]
        return self._data['site_occupancies']

    @property
    def stoichiometry(self):
        """ Return stoichiometry in matador format: a list of
        two-member lists containing element symbol and number
        of atoms per formula unit, sorted in alphabetical order
        by element symbol).

        """
        if 'stoichiometry' not in self._data:
            from matador.utils.chem_utils import get_stoich
            self._data['stoichiometry'] = get_stoich(self.atom_types)
        return self._data['stoichiometry']

    @property
    def concentration(self):
        """ Return concentration of each species in stoichiometry. """
        if 'concentration' not in self._data:
            self._data['concentration'] = get_concentration(
                self.stoichiometry,
                [elem[0] for elem in self.stoichiometry],
                include_end=True
            )
        return self._data['concentration']

    @property
    def formula(self):
        """ Returns chemical formula of structure. """
        from matador.utils.chem_utils import get_formula_from_stoich
        return get_formula_from_stoich(self.stoichiometry, tex=False)

    @property
    def formula_tex(self):
        """ Returns chemical formula of structure in LaTeX format. """
        from matador.utils.chem_utils import get_formula_from_stoich
        return get_formula_from_stoich(self.stoichiometry, tex=True)

    @property
    def cell_volume(self):
        """ Returns cell volume in Å³. """
        return self.cell.volume

    @property
    def lattice_cart(self):
        """ The Cartesian lattice vectors as a tuple. """
        return self.cell.lattice_cart

    @property
    def lattice_abc(self):
        """ Lattice parameters as a tuple. """
        return self.cell.lattice_abc

    @property
    def bond_lengths(self):
        """ Returns a list of ((species_A, species_B), bond_length)),
        sorted by bond length, computed from the network structure of
        the crystal (i.e. first coordination sphere).

        """
        if self._bond_lengths is None:
            self._bond_lengths = []
            for i, j, data in self.network.edges.data():
                self._bond_lengths.append(((self[i].species, self[j].species), data['dist']))
            self._bond_lengths = sorted(self._bond_lengths, key=lambda bond: bond[1])
        return self._bond_lengths

    @property
    def voronoi_substructure(self):
        """ Returns, or calculates  if not present, the Voronoi
        substructure of crystal.

        """
        if 'voronoi_substructure' not in self._data:
            from matador.plugins.voronoi_interface.voronoi_interface import get_voronoi_substructure
            self._data['voronoi_substructure'] = get_voronoi_substructure(self._data)
        return self._data['voronoi_substructure']

    @property
    def space_group(self):
        """ Return the space group symbol at the last-used symprec. """
        return self.get_space_group(symprec=self._data.get('symprec', 0.01))

    def get_space_group(self, symprec=0.01):
        """ Return the space group of the structure at the desired
        symprec. Stores the space group in a dictionary
        `self._space_group` under symprec keys. Updates
        `self._data['space_group']` and `self._data['symprec']` with the
        last value calculated.

        Keyword arguments:
            symprec (float): spglib symmetry tolerance.

        """
        if symprec not in self._space_group:
            self._data['space_group'] = cell_utils.get_spacegroup_spg(self._data, symprec=symprec, check_occ=False)
            self._data['symprec'] = symprec
            self._space_group[symprec] = self._data['space_group']
        return self._space_group[symprec]

    @property
    def pdf(self):
        """ Returns a PDF object (pair distribution function) for the
        structure, calculated with default PDF settings.

        """
        from matador.fingerprints.pdf import PDF
        if 'pdf' not in self._data:
            self._data['pdf'] = PDF(self._data, label=self.formula_tex)
        return self._data['pdf']

    @pdf.setter
    def pdf(self, pdf):
        """ Set the PDF to the given PDF object (or None). """
        from matador.fingerprints.pdf import PDF
        if isinstance(pdf, PDF) or pdf is None:
            self._data['pdf'] = pdf

    def calculate_pdf(self, **kwargs):
        """ Calculate and set the PDF with the passed parameters. """
        from matador.fingerprints.pdf import PDF
        if 'pdf' not in self._data:
            self._data['pdf'] = PDF(self._data, label=self.formula_tex, **kwargs)
        return self._data['pdf']

    @property
    def pxrd(self):
        """ Returns a PXRD object (powder xray diffraction) containing
        the XRD pattern for the structure.

        """
        from matador.fingerprints.pxrd import PXRD
        if 'pxrd' not in self._data:
            self._data['pxrd'] = PXRD(self._data)
        return self._data['pxrd']

    @pxrd.setter
    def pxrd(self, pxrd):
        """ Set the PXRD to the given PXRD object (or None). """
        from matador.fingerprints.pxrd import PXRD
        if isinstance(pxrd, PXRD) or pxrd is None:
            self._data['pxrd'] = pxrd

    def calculate_pxrd(self, **kwargs):
        """ Compute and set PXRD with the passed parameters. """
        from matador.fingerprints.pxrd import PXRD
        if 'pxrd' not in self._data:
            self._data['pxrd'] = PXRD(self._data, **kwargs)
        return self._data['pxrd']

    @property
    def coordination_stats(self):
        """ Returns stastistics on the coordination of each site, as
        computed from Voronoi decomposition.

        """
        if self._coordination_stats is None:
            coordination_lists = self.coordination_lists
            import numpy as np
            coordination_stats = {}
            for species in self.elems:
                coordination_stats[species] = {}
                for _species in self.elems:
                    coordination_stats[species][_species] = {}
                    coordination_stats[species][_species]['mean'] = np.mean(coordination_lists[species][_species])
                    coordination_stats[species][_species]['median'] = np.median(coordination_lists[species][_species])
                    coordination_stats[species][_species]['std'] = np.std(coordination_lists[species][_species])
            self._coordination_stats = coordination_stats

        return self._coordination_stats

    @property
    def ase_atoms(self):
        """ Returns an ASE Atoms representation of
        the crystal.

        """
        from matador.utils.ase_utils import doc2ase
        return doc2ase(self, add_keys_to_info=False)

    @property
    def pmg_structure(self):
        """ Returns the pymatgen structure representation of
        the crystal.

        """
        from matador.utils.pmg_utils import doc2pmg
        return doc2pmg(self)

    @property
    def coordination_lists(self):
        """ Returns a dictionary containing pairs of elements and the
        list of coordination numbers per site.

        """
        if self._coordination_lists is None:
            coordination_lists = {}
            for species in self.elems:
                coordination_lists[species] = {}
                for _species in self.elems:
                    coordination_lists[species][_species] = []
            for site in self:
                for _species in self.elems:
                    if _species in site.coordination:
                        coordination_lists[site.species][_species].append(site.coordination[_species])
                    else:
                        coordination_lists[site.species][_species].append(0)

            self._coordination_lists = coordination_lists

        return self._coordination_lists

    @property
    def unique_sites(self):
        """ Return unique sites using Voronoi decomposition. """
        from matador.plugins.voronoi_interface.voronoi_similarity import get_unique_sites
        get_unique_sites(self._data)
        return self._data['similar_sites']

    @property
    def network(self):
        """ Returns/constructs a CrystalGraph object of the structure. """
        from matador.crystal.network import CrystalGraph
        if self._network is None:
            self._network = CrystalGraph(self, **self._network_kwargs)
        return self._network

    @property
    def network_stats(self):
        """ Return network-calculated coordnation stats. """
        from collections import defaultdict
        coordination = defaultdict(list)
        for node, data in self.network.nodes.data():
            num_edges = len(self.network.edges(node))
            coordination[data['species']].append(num_edges)
        return coordination

    @property
    def bonding_stats(self):
        """ Return network-calculated bonding stats.

        Returns:
            dict: sorted dictionary with root atom as keys and bond
                information as values.

        """
        if self._bonding_stats is None:
            from collections import defaultdict
            bonding_dict = defaultdict(dict)
            network = self.network
            for node in network.nodes:
                bonding_dict[node] = {'species': self.sites[node].species, 'position': self.sites[node].coords, 'bonds': []}
            bonds = set()
            for data in network.edges.data():
                atom_1 = data[0]
                atom_2 = data[1]
                pair = tuple(sorted([atom_1, atom_2]))
                if pair in bonds:
                    continue
                else:
                    bonds.add(pair)

                site_1 = self.sites[atom_1]
                site_2 = self.sites[atom_2]

                bond_length = data[2]['dist']
                is_image = bool(data[2]['image'])
                bonding_dict[atom_1]['bonds'].append(
                    {'species': site_2.species, 'index': atom_2, 'length': bond_length,
                     'is_image': is_image, 'position': site_2.coords}
                )
                bonding_dict[atom_2]['bonds'].append(
                    {'species': site_1.species, 'index': atom_1, 'length': bond_length,
                     'is_image': is_image, 'position': site_1.coords}
                )

            for key in bonding_dict:
                bonding_dict[key]['bonds'] = sorted(bonding_dict[key]['bonds'], key=lambda x: x['index'])

            self._bonding_stats = {key: bonding_dict[key] for key in sorted(bonding_dict)}

        return self._bonding_stats

    def draw_network(self, layout=None):
        """ Draw the CrystalGraph network. """
        from matador.crystal.network import draw_network
        draw_network(self, layout=layout)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements plotting routines specifically
for the PXRD objects defined in the
matador.fingerprints.pxrd module.

"""


from matador.plotting.plotting import plotting_function
from matador.utils.cell_utils import get_space_group_label_latex
from matador.crystal import Crystal


__all__ = ['plot_pxrd']


@plotting_function
def plot_pxrd(
    pxrds, two_theta_range=None, rug=False, rug_height=0.05, rug_offset=0.04, offset=None,
    ax=None, labels=None, figsize=None, text_offset=0.1, filename=None, **kwargs
):
    """ Plot PXRD or PXRDs.

    Parameters:
        pxrds (list or matador.fingerprints.pxrd.PXRD): the PXRD
            or list of PXRDs to plot.

    Keyword arguments:
        two_theta_range (tuple): plotting limits for 2theta
        rug (bool): whether to provide a rug plot of all peaks.
        rug_height (float): size of rug ticks.
        rug_offset (float): offset of rug ticks.
        offset (float): extra space added between patterns (as fraction
            of max intensity). Default 0.1.
        labels (list of str): list of labels to plot alongside pattern.
        figsize (tuple): specify a figure size, the default
            scales with the number of PXRDs to be plotted.
        ax (matplotlib.Axis): optional axis object to plot on.
        text_offset (float): amount by which to offset the labels.
        filename (str): optional filename for saving.


    """

    import matplotlib.pyplot as plt

    if not isinstance(pxrds, list):
        pxrds = [pxrds]
    if labels is not None and not isinstance(labels, list):
        labels = [labels]

    if figsize is None:
        _user_default_figsize = plt.rcParams.get('figure.figsize', (8, 6))
        height = len(pxrds) * max(0.5, _user_default_figsize[1] / 1.5 / len(pxrds))
        figsize = (_user_default_figsize[0], height)

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    if offset is None and rug:
        offset = 0.2
    elif offset is None:
        offset = 0.1

    colour_cycle = ax._get_lines.prop_cycler

    for ind, pxrd in enumerate(pxrds):
        if isinstance(pxrd, Crystal):
            pxrd = pxrd.pxrd
        elif isinstance(pxrd, dict) and 'pxrd' in pxrd:
            pxrd = pxrd['pxrd']

        c = next(colour_cycle).get('color')

        if labels:
            label = labels[ind]
        else:
            label = get_space_group_label_latex(pxrd.spg) + '-' + pxrd.formula

        ax.plot(pxrd.two_thetas, (1 - offset) * pxrd.pattern + ind, c=c)

        ax.text(0.95, ind+text_offset, label,
                transform=ax.get_yaxis_transform(),
                horizontalalignment='right')

        if rug:
            import numpy as np
            peaks = np.unique(pxrd.peak_positions)
            for peak in peaks:
                ax.plot([peak, peak], [ind-rug_height-rug_offset, ind-rug_offset], c=c, alpha=0.5)

    if len(pxrds) > 1:
        ax.set_yticks([])
    else:
        import numpy as np
        ax.set_yticks(np.linspace(0, 1, 5, endpoint=True))
    ax.set_ylim(-0.2, len(pxrds)+0.1)
    if two_theta_range is not None:
        ax.set_xlim(*two_theta_range)
    ax.set_ylabel('Relative intensity')
    ax.set_xlabel('$2\\theta$ (degrees)')

    if any([kwargs.get('pdf'), kwargs.get('svg'), kwargs.get('png')]):
        bbox_extra_artists = None
        if filename is None:
            filename = '-'.join([pxrd.formula for pxrd in pxrds]) + '_pxrd'

        if kwargs.get('pdf'):
            plt.savefig('{}.pdf'.format(filename),
                        bbox_inches='tight', transparent=True, bbox_extra_artists=bbox_extra_artists)
        if kwargs.get('svg'):
            plt.savefig('{}.svg'.format(filename),
                        bbox_inches='tight', transparent=True, bbox_extra_artists=bbox_extra_artists)
        if kwargs.get('png'):
            plt.savefig('{}.png'.format(filename),
                        bbox_inches='tight', transparent=True, bbox_extra_artists=bbox_extra_artists)



# coding: utf-8
# Distributed under the terms of the MIT license.

""" This file implements the BatchRun class for chaining
ComputeTask instances across several structures with
high-throughput.

"""


from collections import defaultdict
import multiprocessing as mp
import os
import glob
import time
import random
import psutil
from matador.utils.print_utils import print_failure, print_warning
from matador.compute.queueing import get_queue_manager
from matador.scrapers.castep_scrapers import cell2dict, param2dict
from matador.compute.compute import ComputeTask
from matador.utils.errors import (
    InputError, CalculationError,
    MaxMemoryEstimateExceeded, NodeCollisionError
)


class BatchRun:
    """ A class that implements the running of multiple generic jobs on
    a series of files without collisions with other nodes using the
    ComputeTask class. Jobs that have been started are listed in
    `jobs.txt`, failed jobs are moved to `bad_castep/`, completed jobs
    are moved to `completed/`.

    Interface initially inspired by on run.pl, run2.pl and PyAIRSS class
    CastepRunner.

    """

    def __init__(self, seed, **kwargs):
        """ Check directory has valid contents and prepare log files
        and directories if not already prepared, then begin running
        calculations.

        Note:
            This class is usually initialised by the run3 script, which
            has a full description of possible arguments.

        Parameters:
            seed (:obj:`list` of :obj:`str`): single entry of param/cell
                file seed for CASTEP geometry optimisations of res
                files, or a list of filenames of $seed to run arbitrary
                executables on. e.g. ['LiAs'] if LiAs.cell and LiAs.param
                exist in cwd full of res files, e.g.2. ['LiAs_1', 'LiAs_2']
                if LiAs_1.in/LiAs_2.in exist, and executable = 'pw6.x < $seed.in'.

        Keyword arguments:
            Exhaustive list found in argparse parser inside `matador/cli/run3.py`.

        """
        # parse args, then co-opt them for passing directly into ComputeTask
        prop_defaults = {'ncores': None, 'nprocesses': 1, 'nnodes': 1,
                         'executable': 'castep', 'no_reopt': False, 'mode': None,
                         'redirect': None, 'debug': False, 'custom_params': False,
                         'verbosity': 0, 'archer': False, 'slurm': False,
                         'intel': False, 'conv_cutoff': False, 'conv_kpt': False,
                         'memcheck': False, 'maxmem': None, 'killcheck': True, 'scratch_prefix': None,
                         'kpts_1D': False, 'spin': None, 'ignore_jobs_file': False,
                         'rough': 4, 'rough_iter': 2, 'fine_iter': 20, 'max_walltime': None,
                         'limit': None, 'profile': False, 'polltime': 30}
        self.args = {}
        self.args.update(prop_defaults)
        self.args.update(kwargs)
        self.debug = self.args.get('debug')

        self.seed = seed
        # if only one seed, check if it is a file, and if so treat
        # this run as a generic run, not a CASTEP cell/param run
        if len(self.seed) == 1 and isinstance(self.seed, list):
            if '*' in self.seed[0]:
                self.seed = glob.glob(self.seed[0])
            elif not os.path.isfile(self.seed[0]):
                self.seed = self.seed[0]

        self.compute_dir = os.uname()[1]
        if self.args.get('scratch_prefix') not in [None, '.']:
            self.compute_dir = '{}/{}'.format(self.args['scratch_prefix'], self.compute_dir).replace('//', '/')
        elif self.args.get('scratch_prefix') == '.':
            self.compute_dir = None

        if self.args.get('mode') is not None:
            self.mode = self.args.get('mode')
        else:
            if isinstance(self.seed, str):
                self.mode = 'castep'
            else:
                self.mode = 'generic'
        del self.args['mode']

        if self.args.get('no_reopt'):
            self.args['reopt'] = False
        else:
            self.args['reopt'] = True
        if 'no_reopt' in self.args:
            del self.args['no_reopt']
        self.nprocesses = int(self.args['nprocesses'])
        del self.args['nprocesses']
        self.limit = self.args.get('limit')
        del self.args['limit']
        self.maxmem = self.args.get('maxmem')
        del self.args['maxmem']
        self.max_walltime = self.args.get('max_walltime')

        # detect and scrape queue settings
        self.queue_mgr = get_queue_manager()

        if self.queue_mgr is not None:
            if self.maxmem is None:
                self.maxmem = self.queue_mgr.max_memory
            if self.max_walltime is None:
                self.max_walltime = self.queue_mgr.walltime

        self.start_time = None
        if self.max_walltime is not None:
            self.start_time = time.time()

        # assign number of cores
        self.all_cores = psutil.cpu_count(logical=False)
        if self.args.get('ncores') is None:
            if self.queue_mgr is None:
                self.args['ncores'] = int(self.all_cores / self.nprocesses)
            else:
                self.args['ncores'] = int(self.queue_mgr.ntasks / self.nprocesses)

        if self.args['nnodes'] < 1 or self.args['ncores'] < 1 or self.nprocesses < 1:
            raise InputError('Invalid number of cores, nodes or processes.')
        if self.all_cores < self.nprocesses:
            raise InputError('Requesting more processes than available cores: {} vs {}'
                             .format(self.all_cores, self.nprocesses))

        # scrape input cell/param/other files
        if self.mode == 'castep':
            self.castep_setup()
        else:
            self.generic_setup()

        # prepare folders and text files
        self.paths = dict()
        if self.args.get('conv_cutoff'):
            self.paths['completed_dir'] = 'completed_cutoff'
        elif self.args.get('conv_kpt'):
            self.paths['completed_dir'] = 'completed_kpts'
        else:
            self.paths['completed_dir'] = 'completed'
        self.paths['failed_dir'] = 'bad_castep'
        self.paths['jobs_fname'] = 'jobs.txt'
        self.paths['completed_fname'] = 'finished_cleanly.txt'
        self.paths['failures_fname'] = 'failures.txt'
        self.paths['memory_fname'] = 'memory_exceeded.txt'
        if not os.path.isfile(self.paths['jobs_fname']):
            with open(self.paths['jobs_fname'], 'a'):
                pass
        if not os.path.isfile(self.paths['completed_fname']):
            with open(self.paths['completed_fname'], 'a'):
                pass
        if not os.path.isfile(self.paths['failures_fname']):
            with open(self.paths['failures_fname'], 'a'):
                pass
        if self.args.get('memcheck'):
            if not os.path.isfile(self.paths['memory_fname']):
                with open(self.paths['memory_fname'], 'a'):
                    pass

    def spawn(self, join=False):
        """ Spawn processes to perform calculations.

        Keyword arguments:
            join (bool): whether or not to attach to ComputeTask
                process. Useful for testing.

        """
        procs = []
        error_queue = mp.Queue()
        for proc_id in range(self.nprocesses):
            procs.append(
                mp.Process(
                    target=self.perform_new_calculations,
                    args=(
                        random.sample(self.file_lists['res'],
                                      len(self.file_lists['res'])),
                        error_queue,
                        proc_id
                    )
                )
            )
        for proc in procs:
            proc.start()
            if join:
                proc.join()

        errors = []
        failed_seeds = []

        # wait for each proc to write to error queue
        try:
            for _, proc in enumerate(procs):
                result = error_queue.get()
                if isinstance(result[1], Exception):
                    errors.append(result)
                    failed_seeds.append(result[2])

            if errors:
                error_message = ''
                for error in errors:
                    error_message += 'Process {} raised error(s): {}. '.format(error[0], error[1])
                    if len({type(error[1]) for error in errors}) == 1:
                        raise errors[0][1]
                    raise type(errors[0][1])(error_message)
                raise BundledErrors(error_message)
        # the only errors that reach here are fatal, e.g. WalltimeError, CriticalError, InputError, KeyboardInterrupt
        except RuntimeError as err:
            result = [proc.join(timeout=2) for proc in procs]
            result = [proc.terminate() for proc in procs if proc.is_alive()]
            print_failure('Fatal error(s) reported:')
            print_warning(err)
            raise err

        print('Nothing left to do.')

    def perform_new_calculations(self, res_list, error_queue, proc_id):
        """ Perform all calculations that have not already
        failed or finished to completion.

        Parameters:
            res_list (:obj:`list` of :obj:`str`): list of structure filenames.
            error_queue (multiprocessing.Queue): queue to push exceptions to
            proc_id (int): process id for logging

        """
        job_count = 0
        if isinstance(res_list, str):
            res_list = [res_list]
        for res in res_list:
            try:
                if not os.path.isfile(res):
                    continue
                # probe once then sleep for a random amount up to 5 seconds
                # before checking again for a lock file, just to protect
                # against collisions in large array jobs on slower parallel file systems
                _ = os.path.isfile('{}.lock'.format(res))
                time.sleep(2 * random.random())
                # wait some additional time if this is a slurm array job
                if self.queue_mgr is not None:
                    extra_wait = self.queue_mgr.array_id % 10 if self.queue_mgr.array_id else 0
                    time.sleep(extra_wait)
                locked = os.path.isfile('{}.lock'.format(res))
                if not self.args.get('ignore_jobs_file'):
                    listed = self._check_jobs_file(res)
                else:
                    listed = []
                running = any([listed, locked])
                if not running:
                    # check we haven't reached job limit
                    if self.limit is not None and job_count >= self.limit:
                        error_queue.put((proc_id, job_count, res))
                        return

                    # check 3 more times if a lock exists with random up to 1 second
                    # waits each time
                    for _ in range(3):
                        time.sleep(random.random())
                        if os.path.isfile('{}.lock'.format(res)):
                            raise NodeCollisionError('Another node wrote this file when I wanted to, skipping...')
                    with open(res + '.lock', 'a') as job_file:
                        pass

                    # write to jobs file
                    with open(self.paths['jobs_fname'], 'a') as job_file:
                        job_file.write(res + '\n')

                    # create full relaxer object for creation and running of job
                    job_count += 1
                    relaxer = ComputeTask(node=None, res=res,
                                          param_dict=self.param_dict,
                                          cell_dict=self.cell_dict,
                                          mode=self.mode, paths=self.paths, compute_dir=self.compute_dir,
                                          timings=(self.max_walltime, self.start_time), maxmem=self.maxmem,
                                          **self.args)
                    # if memory check failed, let other nodes have a go
                    if not relaxer.enough_memory:
                        with open(self.paths['memory_fname'], 'a') as job_file:
                            job_file.write(res + '\n')
                        if os.path.isfile('{}.lock'.format(res)):
                            os.remove('{}.lock'.format(res))
                        with open(self.paths['jobs_fname'], 'r+') as job_file:
                            flines = job_file.readlines()
                            job_file.seek(0)
                        for line in flines:
                            if res not in line:
                                job_file.write(line)
                            job_file.truncate()

                    elif relaxer.success:
                        with open(self.paths['completed_fname'], 'a') as job_file:
                            job_file.write(res + '\n')
                    else:
                        with open(self.paths['failures_fname'], 'a') as job_file:
                            job_file.write(res + '\n')

            # catch memory errors and reset so another node can try
            except MaxMemoryEstimateExceeded:
                reset_single_seed(res)
                continue

            # ignore any other individual calculation errors or node collisions that were caught here
            except CalculationError:
                continue

            # reset txt/lock for an input error, but throw it to prevent other calcs
            except InputError as err:
                reset_single_seed(res)
                error_queue.put((proc_id, err, res))
                return
            # push globally-fatal errors to queue, and return to prevent further calcs
            except RuntimeError as err:
                error_queue.put((proc_id, err, res))
                return
            # finally catch any other generic error in e.g. the above code, normally caused by me
            except Exception as err:
                error_queue.put((proc_id, err, res))
                return

        error_queue.put((proc_id, job_count, ''))

    def generic_setup(self):
        """ Undo things that are set ready for CASTEP jobs... """
        self.cell_dict = None
        self.param_dict = None

        # scan directory for files to run
        self.file_lists = defaultdict(list)
        self.file_lists['res'] = glob.glob('*.res')

    def castep_setup(self):
        """ Set up CASTEP jobs from res files, and $seed.cell/param. """
        # read cell/param files
        exts = ['cell', 'param']
        for ext in exts:
            if not os.path.isfile('{}.{}'.format(self.seed, ext)):
                raise InputError('Failed to find {ext} file, {seed}.{ext}'.format(ext=ext, seed=self.seed))
        self.cell_dict, cell_success = cell2dict(self.seed + '.cell',
                                                 db=False, lattice=False, positions=True)
        if not cell_success:
            print(self.cell_dict)
            raise InputError('Failed to parse cell file')
        self.param_dict, param_success = param2dict(self.seed + '.param', db=False)
        if not param_success:
            print(self.param_dict)
            raise InputError('Failed to parse param file')

        # scan directory for files to run
        self.file_lists = defaultdict(list)
        self.file_lists['res'] = glob.glob('*.res')
        if any(self.seed == file.replace('.res', '') for file in self.file_lists['res']):
            error = ("Found .res file with same name as seed: {}.res. This will wreak havoc on your calculations!\n"
                     .format(self.seed)
                     + "Please rename either your seed.cell/seed.param files, or rename the offending {}.res"
                     .format(self.seed))
            raise InputError(error)

        if not self.file_lists['res']:
            error = (
                'run3 in CASTEP mode requires at least 1 res file in folder, found {}'
                .format(len(self.file_lists['res']))
            )
            raise InputError(error)

        if (len(self.file_lists['res']) < self.nprocesses
                and not any([self.args.get('conv_cutoff'), self.args.get('conv_kpt')])):
            raise InputError('Requested more processes than there are jobs to run!')

        # do some prelim checks of parameters
        if self.param_dict['task'].upper() in ['GEOMETRYOPTIMISATION', 'GEOMETRYOPTIMIZATION']:
            if 'geom_max_iter' not in self.param_dict:
                raise InputError('geom_max_iter is unset, please fix this.')
            if int(self.param_dict['geom_max_iter']) <= 0:
                raise InputError('geom_max_iter is only {}!'.format(self.param_dict['geom_max_iter']))

        # parse convergence args and set them up
        self.convergence_run_setup()

        # delete source from cell and param
        del self.cell_dict['source']
        del self.param_dict['source']

    def convergence_run_setup(self):
        """ Set the correct args for a convergence run. """
        # check if we're doing a conv run
        if self.args.get('conv_cutoff'):
            if os.path.isfile('cutoff.conv'):
                with open('cutoff.conv', 'r') as f:
                    flines = f.readlines()
                    self.args['conv_cutoff'] = []
                    for line in flines:
                        if not line.startswith('#'):
                            self.args['conv_cutoff'].append(int(line))
            else:
                raise InputError('Missing cutoff.conv file')
        else:
            self.args['conv_cutoff'] = None

        if self.args.get('conv_kpt'):
            if os.path.isfile('kpt.conv'):
                with open('kpt.conv', 'r') as f:
                    flines = f.readlines()
                    self.args['conv_kpt'] = []
                    for line in flines:
                        if not line.startswith('#'):
                            self.args['conv_kpt'].append(float(line))
            else:
                raise InputError('Missing with conv.kpt file')
        else:
            self.args['conv_kpt'] = None

    def _check_jobs_file(self, res):
        """ Check if structure is listed in jobs.txt file.

        Parameters:
            res (str): structure name.

        Returns:
            bool: True if already listed in jobs file.

        """
        with open(self.paths['jobs_fname'], 'r') as job_file:
            flines = job_file.readlines()
            for line in flines:
                if res in line:
                    return True
        return False


class BundledErrors(Exception):
    """ Raise this after collecting all exceptions from
    processes.
    """


def reset_job_folder(debug=False):
    """ Remove all lock files and clean up jobs.txt
    ready for job restart.

    Note:
        This should be not called by a ComputeTask instance, in case
        other instances are running.

    Returns:
        num_remaining (int): number of structures left to relax

    """
    res_list = glob.glob('*.res')
    if debug:
        print(res_list)
    for f in res_list:
        root = f.replace('.res', '')
        exts_to_rm = ['res.lock', 'kill']
        for ext in exts_to_rm:
            if os.path.isfile('{}.{}'.format(root, ext)):
                if debug:
                    print('Deleting {}.{}'.format(root, ext))
                os.remove('{}.{}'.format(root, ext))

    # also remove from jobs file
    if os.path.isfile('jobs.txt'):
        with open('jobs.txt', 'r+') as f:
            flines = f.readlines()
            if debug:
                print('Initially {} jobs in jobs.txt'.format(len(flines)))
            f.seek(0)
            for line in flines:
                line = line.strip()
                if line in res_list:
                    print('Excluding {}'.format(line))
                    continue
                f.write(line)
            f.truncate()
            flines = f.readlines()
            if debug:
                print('{} jobs remain in jobs.txt'.format(len(flines)))

    return len(res_list)


def reset_single_seed(seed):
    """ Remove the file lock and jobs.txt entry
    for a single seed.

    Parameters:
        seed (str): the seedname to remove.

    """
    if os.path.isfile('jobs.txt'):
        with open('jobs.txt', 'r+') as f:
            flines = f.readlines()
            f.seek(0)
            for line in flines:
                line = line.strip()
                if seed in line:
                    continue
                f.write(line)
            f.truncate()
            flines = f.readlines()
    if os.path.isfile(seed + '.lock'):
        os.remove(seed + '.lock')



import numpy as np
from matador.utils.cell_utils import real2recip, frac2cart
from matador.orm.orm import DataContainer


class Spectral(DataContainer):
    """
    Note:
        This class speaks of "k-points" as general reciprocal space points
        used to display the dispersion curves; these correspond to CASTEP's
        phonon_kpoints or spectral_kpoints, and not the k-points
        used to generate the underlying wavefunction or dynamical matrix.

    """

    @property
    def eigs(self):
        """ Alias for the correct eigenvalue array. """
        if 'Vibrational' in self.__class__.__name__:
            return self._data['eigs_q']
        return self._data['eigs_s_k']

    @property
    def lattice_cart(self):
        """ The Cartesian lattice vectors of the real space lattice. """
        return self._data['lattice_cart']

    @property
    def num_kpoints(self):
        """ Number of dispersion k-points sampled. """
        return self._data['num_kpoints']

    @property
    def num_qpoints(self):
        """ Alias for number of kpoints. """
        return self.num_kpoints

    @property
    def projectors(self):
        """ Return list of projector labels in the format
        `(element, l-channel)`.

        """
        return self._data.get('projectors')

    @property
    def num_modes(self):
        """ Number of eigenvalues per q/k-point. """
        return self._data['num_modes']

    @property
    def num_bands(self):
        """ Number of eigenvalues per q/k-point. """
        if 'Vibrational' in self.__class__.__name__:
            return self._data['num_modes']
        return self._data['num_bands']

    @property
    def projector_weights(self):
        """ Return the array of projector weights per eigval, with shape
        (num_projectors, num_kpoints, num_bands).

        """
        return self._data.get('projector_weights')

    @property
    def num_projectors(self):
        """ Return the number of projectors. """
        if self.projectors is None:
            return 0
        return len(self.projectors)

    @property
    def kpoint_branches(self):
        """ Return the k-point branches in the older format, which
        contained a list of lists of continous indices.

        """
        if self._data.get('kpoint_branches') is None:
            self._data['kpoint_branches'] = self.find_full_kpt_branch()
        return self._data['kpoint_branches']

    @property
    def kpoint_branch_start(self):
        """ Return the indices of the start of branches. """
        if not self._data.get('kpoint_branch_start'):
            self.set_branches_and_spacing()
        return self._data['kpoint_branch_start']

    @property
    def kpoint_path_spacing(self):
        """ An estimated kpoint spacing. """
        if not self._data.get('kpoint_path_spacing'):
            self.set_branches_and_spacing()
        return self._data['kpoint_path_spacing']

    @property
    def kpoint_path(self):
        """ The fractional sampling path in reciprocal space. """
        return np.asarray(self._data['kpoint_path'])

    @property
    def kpoint_weights(self):
        if 'kpoint_weights' in self._data:
            return np.asarray(self._data['kpoint_weights'])
        return None

    @property
    def kpoint_path_cartesian(self):
        """ The reicprocal space sampling path in Cartesian coordinates. """
        return np.asarray(frac2cart(real2recip(self.lattice_cart),
                                    self.kpoint_path))

    @property
    def num_spins(self):
        """ Dummy number of spins. """
        return 1

    @property
    def spin_fermi_energy(self):
        """ Dummy Fermi energy per spin channel. """
        return [0]



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule provides data loaders for the tabulated atomic
scattering data stored within the module.

"""


def load_scattering_file(filename, a_inds, b_inds, c_ind):
    """ Load a scattering file with a, b and c coefficients
    in the columns provided.

    """

    import numpy as np

    with open(filename, "r") as f:
        flines = [line for line in f.readlines() if not line.startswith("#")]

    atomic_scattering_coeffs = {}

    for line in flines:
        x = line.split()
        label = x[0]
        a = np.array([float(x[ind]) for ind in a_inds])
        b = np.array([float(x[ind]) for ind in b_inds])
        c = float(x[c_ind])
        atomic_scattering_coeffs[label] = [a, b, c]

    return atomic_scattering_coeffs



#!/usr/bin/env python

import unittest

try:
    import ase  # noqa
    import ase.build

    ASE_IMPORTED = True
except ImportError:
    ASE_IMPORTED = False

from matador.utils.ase_utils import ase2dict


@unittest.skipIf(not ASE_IMPORTED, "ASE not found")
class AseUtilTest(unittest.TestCase):
    """ Tests cursor util functions. """

    def setUp(self):
        # construct ASE atoms object manually
        self.ase_atoms = ase.build.bulk(
            "SiO", crystalstructure="rocksalt", a=12, cubic=True
        )

    def test_ase2dict(self):
        doc = ase2dict(self.ase_atoms)
        self.assertEqual(
            doc["atom_types"], sorted(["Si", "O", "Si", "O", "Si", "O", "Si", "O"])
        )
        self.assertListEqual(doc["lattice_cart"][0], [12, 0, 0])
        self.assertListEqual(doc["lattice_cart"][1], [0, 12, 0])
        self.assertListEqual(doc["lattice_cart"][2], [0, 0, 12])
        self.assertListEqual(doc["lattice_abc"][0], [12, 12, 12])
        self.assertListEqual(doc["lattice_abc"][1], [90, 90, 90])
        self.assertAlmostEqual(doc["cell_volume"], 12 ** 3)
        self.assertEqual(len(doc["positions_frac"]), len(doc["atom_types"]))
        self.assertListEqual(doc["positions_frac"][0], [0.5, 0.0, 0.0])
        self.assertListEqual(doc["positions_frac"][1], [0.5, 0.5, 0.5])
        self.assertListEqual(doc["positions_frac"][3], [0.0, 0.5, 0.0])
        self.assertEqual(doc["num_atoms"], 8)
        self.assertEqual(doc["num_fu"], 4)
        self.assertListEqual(doc["stoichiometry"][0], ["O", 1.0])
        self.assertListEqual(doc["stoichiometry"][1], ["Si", 1.0])
        self.assertListEqual(sorted(list(doc["elems"])), ["O", "Si"])
        self.assertEqual(doc["space_group"], "Fm-3m")

    def test_doc2ase(self):
        pass


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import unittest
import numpy as np

from matador.scrapers import castep2dict
from matador.orm.spectral import (
    VibrationalDOS,
    VibrationalDispersion,
    ElectronicDispersion,
    ElectronicDOS,
)
from .utils import REAL_PATH


class SpectralOrmTest(unittest.TestCase):
    def test_failed_make_from_castep(self):
        """ Tests that an error is raised when trying to create
        a Spectral object from a castep file that contains no
        spectral data.

        """
        fname = REAL_PATH + "data/castep_files/KP-castep17.castep"
        doc, s = castep2dict(fname, db=False)

        with self.assertRaises(RuntimeError):
            ElectronicDispersion(doc)

        with self.assertRaises(RuntimeError):
            ElectronicDispersion(**doc)

        with self.assertRaises(RuntimeError):
            ElectronicDOS(doc)

        with self.assertRaises(RuntimeError):
            ElectronicDOS(**doc)

        with self.assertRaises(RuntimeError):
            VibrationalDOS(doc)

        with self.assertRaises(RuntimeError):
            VibrationalDOS(**doc)

        with self.assertRaises(RuntimeError):
            VibrationalDispersion(doc)

        with self.assertRaises(RuntimeError):
            VibrationalDispersion(**doc)

    def test_dos_construction_from_frequencies(self):
        """ Test that we can make sensible phonon DOS's and free
        energy calculations from frequency and qpoint data, and that
        it agrees with CASTEP's calculated values.

        """

        fname = REAL_PATH + "data/castep_files/CuP-thermo-test.castep"
        doc, _ = castep2dict(fname, db=False)
        dos = VibrationalDOS(doc)
        backup_eigs = np.copy(dos.eigs)
        self.assertAlmostEqual(
            dos.zpe, doc["thermo_zero_point_energy"] / doc["num_atoms"], places=4
        )
        np.testing.assert_array_equal(
            dos.eigs, backup_eigs, err_msg="Eigenvalues were changed by some function"
        )
        self.assertAlmostEqual(
            dos.vibrational_free_energy(1000),
            doc["thermo_free_energy"][1000.0] / doc["num_atoms"],
            places=4,
        )
        np.testing.assert_array_equal(
            dos.eigs, backup_eigs, err_msg="Eigenvalues were changed by some function"
        )
        self.assertEqual(dos.compute_free_energy(0.0), dos.zpe)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" The config simply loads the default or user-specified config. """


__all__ = ['load_custom_settings', 'set_settings', 'SETTINGS']
__author__ = 'Matthew Evans'
__maintainer__ = 'Matthew Evans'

from matador.config.config import load_custom_settings, SETTINGS, set_settings



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule contains functionality to fit E(V) equations
of state to ab initio data, primarily to calculate bulk moduli.

"""

import re

import numpy as np
from matador.utils.chem_utils import eV_PER_ANGSTROM_CUBED_TO_GPa


def get_equation_of_state(seed, plot=False):
    """ Extract E(V) data from CASTEP files and perform
    fits for the equation of state and bulk moduli.

    Parameters:
        seed (str/dict): filename or scraped dictionary to fit.

    Keyword arguments:
        plot (bool): plot the fitted EoS.

    """
    if isinstance(seed, dict):
        results = seed
    else:
        from matador.scrapers import castep2dict
        results, success = castep2dict(seed, intermediates=True, db=False)
        if not success:
            raise RuntimeError(results)

    volumes = []
    energies = []
    for snapshot in results['intermediates']:
        volumes.append(snapshot['cell_volume'])
        energies.append(snapshot['total_energy'])
    volumes.append(results['cell_volume'])
    energies.append(results['total_energy'])
    volumes = np.asarray(volumes)
    energies = np.asarray(energies)

    if len(volumes) < 3:
        raise RuntimeError('Seed {} does not contain enough energies vs volumes to fit a bulk modulus')

    energies = energies[np.argsort(volumes)]
    volumes = np.sort(volumes)

    E_0 = np.min(energies)
    V_0 = volumes[np.argmin(energies)]

    types_of_fit = EquationOfState.__subclasses__()

    results['eos'] = []
    results['summary'] = []
    probes_all = []
    curves = []
    labels = []

    for eos_type in types_of_fit:
        eos = eos_type(E_0, V_0)
        eos.fit(volumes, energies)

        results['summary'].append(eos_type.__name__ + '\n')
        results['summary'].append('fitting parameters = {d[0]:.6f}, {d[1]:.6f}\n'.format(d=eos.fit_parameters))
        results['summary'].append('rrrmsd = {:10.10f} %\n'.format(eos.rrmsd))
        results['summary'].append('bulk modulus = {d[0]:.6f} +/- {d[1]:.6f} GPa\n'
                                  .format(d=(eos.bulk_modulus, eos.bulk_modulus_err)))
        results['summary'].append(80*'-' + '\n')
        probes = np.linspace(min(volumes), max(volumes), num=100)
        fitted_curve = eos.evaluate(probes, *eos.popt)
        results['eos'].append(eos)
        probes_all.append(probes)
        curves.append(fitted_curve)
        name = eos_type.__name__.replace('EOS', '')
        label = '-'.join(re.findall('[A-Z][^A-Z]*', name))
        labels.append(label)

    if plot:
        import matplotlib.pyplot as plt
        _, ax = plt.subplots(1, figsize=(8, 6))
        ax.set_ylabel('Total energy (eV)')
        ax.set_xlabel('Cell volume ($\\AA^3$)')
        ls = ['--', '-.', ':']
        for ind, probe in enumerate(probes_all):
            ax.plot(probes_all[ind], curves[ind], label=labels[ind], ls=ls[ind])
        ax.plot(volumes, energies, marker='o', label='DFT data')

        ax.legend(loc=0)
        plt.savefig(seed + '.png')

    return results


class EquationOfState:
    """ Abstract class for E(V) isothermal equations of state. Used to
    perform least squares fitting to arbitrary functional form.

    Attributes:
        E_0 (float): equilbirium lattice energy.
        V_0 (float): equilibrium lattice volume.
        popt (:obj:`list` of :obj:`float`): fitting parameters for
            particular EOS.
        p0 (:obj:`list` of :obj:`float`): initial guess at fitting parameters.
        rrmsd (float): relative root mean square deviation of fit, as
            defined by [2].

    """

    def __init__(self, E_0, V_0):
        """ Set up EOS ready for fit.

        Parameters:
            E_0 (float): equilibrium energy.
            V_0 (float): equilibrium volume.

        """
        self.E_0 = E_0
        self.V_0 = V_0
        self.popt = None
        self.rrmsd = None
        self.p0 = [0.01, 1]

    def fit(self, volumes, energies):
        """ Perform a least squares fit on the volumes and energies.

        Parameters:
            volumes (:obj:`list` of :obj:`float`): calculated volumes.
            energies (:obj:`list` of :obj:`float`): calculated energies.

        """
        from scipy import optimize
        self.popt, _ = optimize.leastsq(self.residual, self.p0, args=(volumes, energies))
        self.rrmsd = np.sqrt(np.sum(((energies - self.evaluate(volumes, *self.popt)) / energies)**2) / (len(energies) - 1))

    def evaluate(self, V, B, C):
        """ Evaluate the EOS.

        Parameters:
            V (:obj:`list` of :obj:`float`): volumes at which to test.
            B (float): the first fitting parameter defined in [2].
            C (float): the second fitting parameter defined in [2].

        """
        raise NotImplementedError

    def residual(self, guess, V, E):
        """ Calculate the resdiual of the current fit.

        Parameters:
            guess (:obj:`list` of :obj:`float`): interim fitting parameters.
            V (:obj:`list` of :obj:`float`): volumes at which to test.
            E (:obj:`list` of :obj:`float`): energies at the volumes V.

        """
        return E - self.evaluate(V, *guess)

    @property
    def bulk_modulus(self):
        """ Returns the bulk modulus predicted by the fit. """
        bulk_modulus = self.get_bulk_modulus()
        return bulk_modulus

    @property
    def bulk_modulus_err(self):
        """ Returns the estimated error on the bulk modulus. """
        return self.bulk_modulus * self.rrmsd

    @property
    def fit_parameters(self):
        """ Return list of final fitting parameters. """
        return self.popt


class BirchMurnaghanEulerianEOS(EquationOfState):
    """ Implements the 3rd order Birch-Murnaghan EOS [1] for given data, as
    provided by Ref. [2] in the Eulerian frame.

    [1]. Francis Birch, Phys. Rev. 71 809 (1947)
         DOI: 10.1103/PhysRev.71.809.

    [2]. K. Latimer, S. Dwaraknath, K. Mathew, D. Winston, K. A. Persson,
         npj Comput. Mater. 2018 41 2018, 4, 40.
         DOI: 10.1038/s41524-018-0091-x.

    """

    def evaluate(self, V, B, C):
        nu = V / self.V_0
        nu = (nu**(-2.0/3.0) - 1)
        return self.E_0 + B * self.V_0 * (nu**2 + 0.5 * C * nu**3)

    def get_bulk_modulus(self):
        """ Return the bulk modulus of this particular fit. """
        if self.popt is None:
            raise RuntimeError('No fit performed.')
        return self.popt[0] * 8.0/9.0 * eV_PER_ANGSTROM_CUBED_TO_GPa


class PoirerTarantolaEOS(EquationOfState):
    """ Implements the logarithmic Poirer-Tarantola [3] for given data, as
    provided by Ref. [2].

    [2]. K. Latimer, S. Dwaraknath, K. Mathew, D. Winston, K. A. Persson,
         npj Comput. Mater. 2018 41 2018, 4, 40.
         DOI: 10.1038/s41524-018-0091-x.

    [3]. J. Poirer, A. Tarantola, Phys. Earth Planet. Inter. 109 1-8 (1998).

    """

    def evaluate(self, V, B, C):
        nu = V / self.V_0
        return self.E_0 + B * self.V_0 * np.log(nu)**2 * (3 - C*np.log(nu))

    def get_bulk_modulus(self):
        """ Return the bulk modulus of this particular fit. """
        if self.popt is None:
            raise RuntimeError('No fit performed.')
        return self.popt[0] * 6 * eV_PER_ANGSTROM_CUBED_TO_GPa


class TaitEOS(EquationOfState):
    """ Implements the exponential Tait EOS [4] for given data, as
    provided by Ref. [2].

    [2]. K. Latimer, S. Dwaraknath, K. Mathew, D. Winston, K. A. Persson,
         npj Comput. Mater. 2018 41 2018, 4, 40.
         DOI: 10.1038/s41524-018-0091-x.

    [4]. J. H. Dymond, R. Malhotra, Int. J. Thermophys. 9 941-951 (1988).

    """

    def evaluate(self, V, B, C):
        nu = V / self.V_0
        return self.E_0 + (B*self.V_0 / C) * (nu - 1 + (1/C) * (np.exp(C*(1-nu)) - 1))

    def get_bulk_modulus(self):
        """ Return the bulk modulus of this particular fit. """
        if self.popt is None:
            raise RuntimeError('No fit performed.')
        return self.popt[0] * eV_PER_ANGSTROM_CUBED_TO_GPa
